import ubelt as ub
import shlex
import json
# import watch
import shlex  # NOQA
import json  # NOQA
from watch.utils.lightning_ext import util_globals  # NOQA
import kwarray  # NOQA
import networkx as nx
import itertools as it
from watch.utils import util_dotdict


def resolve_pipeline_row(grid_item_defaults, state, region_model_dpath, expt_dvc_dpath, item):
    """
    The idea is to handle any dependencies between parameters, extract any
    path-based metadata, and construct every path that this pipeline will
    touch.
    """
    from watch.mlops.expt_manager import ExperimentState
    state = ExperimentState(expt_dvc_dpath, '*')
    item = grid_item_defaults | item
    # print('item = {}'.format(ub.urepr(item, nl=1)))
    nested = util_dotdict.dotdict_to_nested(item)

    condensed = {}
    paths = {}

    # Might not need this exactly
    import xdev
    with xdev.embed_on_exception_context:
        try:
            pkg_trk_pixel_pathcfg = state._parse_pattern_attrs(
                state.templates['pkg_trk_pxl_fpath'], item['trk.pxl.package_fpath'])
            pkg_trk_pixel_pathcfg.pop('dataset_code')
            pkg_trk_pixel_pathcfg.pop('storage_dpath')
        except RuntimeError:
            ...  # user specified a custom package_fpath
            pkg_trk_pixel_pathcfg = {
                'trk_expt': 'trk_expt_unknown',
                'trk_model': ub.Path(item['trk.pxl.package_fpath']).name,
            }
        condensed.update(pkg_trk_pixel_pathcfg)

    condensed['storage_dpath'] = state.storage_dpath
    condensed['expt_dvc_dpath'] = expt_dvc_dpath

    # might ultimately not be needed.
    condensed['dataset_code'] = 'fixme'

    ### BAS / TRACKING ###

    trk_pxl  = nested['trk']['pxl']
    trk_poly = nested['trk']['poly']
    trk_pxl_params = ub.udict(trk_pxl['data']) - {'test_dataset'}
    trk_poly_params = ub.udict(trk_poly)

    # condensed['trk_model'] = state._condense_model(item['trk.pxl.package_fpath'])
    # TODO:
    # based on the package_fpath, we should infer if we need team features or not.

    condensed['test_trk_dset'] = state._condense_test_dset(item['trk.pxl.data.test_dataset'])
    condensed['trk_pxl_cfg'] = state._condense_cfg(trk_pxl_params, 'trk_pxl')
    condensed['trk_poly_cfg'] = state._condense_cfg(trk_poly_params, 'trk_poly')

    paths['pkg_trk_pxl_fpath'] = ub.Path(item['trk.pxl.package_fpath'])
    paths['trk_test_dataset_fpath'] = item['trk.pxl.data.test_dataset']
    paths['pred_trk_pxl_dpath'] = ub.Path(state.templates['pred_trk_pxl_dpath'].format(**condensed))
    paths['pred_trk_pxl_fpath'] = ub.Path(state.templates['pred_trk_pxl_fpath'].format(**condensed))
    paths['eval_trk_pxl_dpath'] = ub.Path(state.templates['eval_trk_pxl_dpath'].format(**condensed))
    paths['eval_trk_pxl_fpath'] = ub.Path(state.templates['eval_trk_pxl_fpath'].format(**condensed))

    paths['pred_trk_poly_sites_fpath'] = ub.Path(state.templates['pred_trk_poly_sites_fpath'].format(**condensed))
    paths['pred_trk_poly_site_summaries_fpath'] = ub.Path(state.templates['pred_trk_poly_site_summaries_fpath'].format(**condensed))
    paths['pred_trk_poly_sites_dpath'] = ub.Path(state.templates['pred_trk_poly_sites_dpath'].format(**condensed))
    paths['pred_trk_poly_site_summaries_dpath'] = ub.Path(state.templates['pred_trk_poly_site_summaries_dpath'].format(**condensed))

    paths['pred_trk_poly_viz_stamp'] = ub.Path(state.templates['pred_trk_poly_viz_stamp'].format(**condensed))

    paths['pred_trk_poly_dpath'] = ub.Path(state.templates['pred_trk_poly_dpath'].format(**condensed))
    paths['pred_trk_poly_kwcoco'] = ub.Path(state.templates['pred_trk_poly_kwcoco'].format(**condensed))

    paths['eval_trk_poly_fpath'] = ub.Path(state.templates['eval_trk_poly_fpath'].format(**condensed))
    paths['eval_trk_poly_dpath'] = ub.Path(state.templates['eval_trk_poly_dpath'].format(**condensed))

    trackid_deps = {}
    trackid_deps = ub.udict(condensed) & state.hashid_dependencies['trk_poly_id']
    condensed['trk_poly_id'] = state._condense_cfg(trackid_deps, 'trk_poly_id')

    ### CROPPING ###

    crop_params = ub.udict(nested['crop']).copy()
    crop_src_fpath = crop_params.pop('src')
    paths['crop_src_fpath'] = crop_src_fpath
    if crop_params['regions'] == 'truth':
        # Crop job depends only on true annotations
        paths['crop_regions'] = str(region_model_dpath) + '/*.geojson'
        condensed['regions_id'] = 'truth'  # todo: version info
        site_summary = str(region_model_dpath) + '/*.geojson'
    if crop_params['regions'] == 'trk.poly.output':
        # Crop job depends on track predictions
        paths['crop_regions'] = paths['pred_trk_poly_site_summaries_fpath']
        condensed['regions_id'] = condensed['trk_poly_id']
        site_summary = state.templates['pred_trk_poly_site_summaries_fpath'].format(**condensed)

    crop_params['regions'] = paths['crop_regions']
    condensed['crop_cfg'] = state._condense_cfg(crop_params, 'crop')
    condensed['crop_src_dset'] = state._condense_test_dset(crop_src_fpath)

    crop_id_deps = ub.udict(condensed) & state.hashid_dependencies['crop_id']
    condensed['crop_id'] = state._condense_cfg(crop_id_deps, 'crop_id')
    paths['crop_dpath'] = ub.Path(state.templates['crop_dpath'].format(**condensed))
    paths['crop_fpath'] = ub.Path(state.templates['crop_fpath'].format(**condensed))
    condensed['crop_dst_dset'] = state._condense_test_dset(paths['crop_fpath'])

    ### SC / ACTIVITY ###
    act_pxl  = nested['act']['pxl']
    act_poly = nested['act']['poly']
    act_pxl_params = ub.udict(act_pxl['data']) - {'test_dataset'}
    act_poly_params = ub.udict(act_poly)
    act_poly_params['site_summary'] = site_summary

    condensed['act_model'] = state._condense_model(item['act.pxl.package_fpath'])
    # TODO:
    # based on the package_fpath, we should infer if we need team features or not.
    condensed['act_pxl_cfg'] = state._condense_cfg(act_pxl_params, 'act_pxl')
    condensed['act_poly_cfg'] = state._condense_cfg(act_poly_params, 'act_poly')

    # try:
    #     pkg_act_pixel_pathcfg = state._parse_pattern_attrs(state.templates['pkg_act_pxl_fpath'], item['act.pxl.package_fpath'])
    # except RuntimeError:
    #     ...  # user specified a custom package_fpath
    #     condensed['dataset_code'] = 'dset_code_unknown'
    #     condensed['act_expt'] = 'act_expt_unknown'
    #     condensed['act_model'] = ub.Path(item['act.pxl.package_fpath']).name
    # else:
    #     condensed.update(pkg_act_pixel_pathcfg)

    try:
        pkg_act_pixel_pathcfg = state._parse_pattern_attrs(
            state.templates['pkg_act_pxl_fpath'], item['act.pxl.package_fpath'])
        pkg_act_pixel_pathcfg.pop('dataset_code')
        pkg_act_pixel_pathcfg.pop('storage_dpath')
    except RuntimeError:
        ...  # user specified a custom package_fpath
        pkg_act_pixel_pathcfg = {
            'act_expt': 'act_expt_unknown',
            'act_model': ub.Path(item['act.pxl.package_fpath']).name,
        }
    condensed.update(pkg_act_pixel_pathcfg)

    # paths['act_test_dataset_fpath'] = item['act.pxl.data.test_dataset']
    if item['act.pxl.data.test_dataset'] == 'crop.dst':
        # Activity prediction depends on a cropping job
        paths['act_test_dataset_fpath'] = paths['crop_fpath']
    else:
        # Activity prediction has no dependencies in this case.
        paths['act_test_dataset_fpath'] = item['act.pxl.data.test_dataset']
    condensed['test_act_dset'] = state._condense_test_dset(paths['act_test_dataset_fpath'])

    paths['pkg_act_pxl_fpath'] = ub.Path(item['act.pxl.package_fpath'])

    paths['pred_act_poly_sites_fpath'] = ub.Path(state.templates['pred_act_poly_sites_fpath'].format(**condensed))
    # paths['pred_act_poly_site_summaries_fpath'] = ub.Path(state.templates['pred_act_poly_site_summaries_fpath'].format(**condensed))
    paths['pred_act_poly_sites_dpath'] = ub.Path(state.templates['pred_act_poly_sites_dpath'].format(**condensed))
    # paths['pred_act_poly_site_summaries_dpath'] = ub.Path(state.templates['pred_act_poly_site_summaries_dpath'].format(**condensed))

    paths['pred_act_pxl_dpath'] = ub.Path(state.templates['pred_act_pxl_dpath'].format(**condensed))
    paths['pred_act_poly_dpath'] = ub.Path(state.templates['pred_act_poly_dpath'].format(**condensed))
    paths['pred_act_pxl_fpath'] = ub.Path(state.templates['pred_act_pxl_fpath'].format(**condensed))
    paths['eval_act_pxl_fpath'] = ub.Path(state.templates['eval_act_pxl_fpath'].format(**condensed))
    paths['eval_act_pxl_dpath'] = ub.Path(state.templates['eval_act_pxl_dpath'].format(**condensed))
    paths['eval_act_poly_fpath'] = ub.Path(state.templates['eval_act_poly_fpath'].format(**condensed))
    paths['eval_act_poly_dpath'] = ub.Path(state.templates['eval_act_poly_dpath'].format(**condensed))
    paths['pred_act_poly_kwcoco'] = ub.Path(state.templates['pred_act_poly_kwcoco'].format(**condensed))
    paths['pred_act_poly_viz_stamp'] = ub.Path(state.templates['pred_act_poly_viz_stamp'].format(**condensed))

    # paths['eval_act_tmp_dpath'] = pred_act_row['eval_act_poly_dpath'] / '_tmp'

    task_params = {
        'trk.pxl': trk_pxl_params,
        'trk.poly': trk_poly_params,
        'crop': crop_params,
        'act.pxl': act_pxl_params,
        'act.poly': act_poly_params,
    }

    row = {
        'condensed': condensed,
        'paths': paths,
        'item': item,
        'task_params': task_params,
    }
    return row


def resolve_package_paths(model_globstr, expt_dvc_dpath):
    import rich
    # import glob
    from watch.utils import util_pattern

    # HACK FOR DVC PTH FIXME:
    # if str(model_globstr).endswith('.txt'):
    #     from watch.utils.simple_dvc import SimpleDVC
    #     print('model_globstr = {!r}'.format(model_globstr))
    #     # if expt_dvc_dpath is None:
    #     #     expt_dvc_dpath = SimpleDVC.find_root(ub.Path(model_globstr))

    def expand_model_list_file(model_lists_fpath, expt_dvc_dpath=None):
        """
        Given a file containing paths to models, expand it into individual
        paths.
        """
        expanded_fpaths = []
        lines = [line for line in ub.Path(model_globstr).read_text().split('\n') if line]
        missing = []
        for line in lines:
            if expt_dvc_dpath is not None:
                package_fpath = ub.Path(expt_dvc_dpath / line)
            else:
                package_fpath = ub.Path(line)
            if package_fpath.is_file():
                expanded_fpaths.append(package_fpath)
            else:
                missing.append(line)
        if missing:
            rich.print('[yellow] WARNING: missing = {}'.format(ub.urepr(missing, nl=1)))
            rich.print(f'[yellow] WARNING: specified a models-of-interest.txt and {len(missing)} / {len(lines)} models were missing')
        return expanded_fpaths

    print('model_globstr = {!r}'.format(model_globstr))
    model_globstr = util_pattern.MultiPattern.coerce(model_globstr)
    package_fpaths = []
    # for package_fpath in glob.glob(model_globstr, recursive=True):
    for package_fpath in model_globstr.paths(recursive=True):
        package_fpath = ub.Path(package_fpath)
        if package_fpath.name.endswith('.txt'):
            # HACK FOR PATH OF MODELS
            model_lists_fpath = package_fpath
            expanded_fpaths = expand_model_list_file(model_lists_fpath, expt_dvc_dpath=expt_dvc_dpath)
            package_fpaths.extend(expanded_fpaths)
        else:
            package_fpaths.append(package_fpath)

    if len(package_fpaths) == 0:
        import pathlib
        if '*' not in str(model_globstr):
            package_fpaths = [ub.Path(model_globstr)]
        elif isinstance(model_globstr, (str, pathlib.Path)):
            # Warn the user if they gave a bad model globstr (this is just one
            # of the many potential ways things could go wrong)
            glob_path = ub.Path(model_globstr)

            def _concrete_glob_part(path):
                " Find the resolved part of the glob path "
                concrete_parts = []
                for p in path.parts:
                    if '*' in p:
                        break
                    concrete_parts.append(p)
                return ub.Path(*concrete_parts)
            concrete = _concrete_glob_part(glob_path)
            if not concrete.exists():
                rich.print('[yellow] WARNING: part of the model_globstr does not exist: {}'.format(concrete))

    return package_fpaths


class Step:
    def __init__(step, name, command, in_paths, out_paths, resources):
        step.name = name
        step._command = command
        step.in_paths = ub.udict(in_paths)
        step.out_paths = ub.udict(out_paths)
        step.resources = ub.udict(resources)
        #
        # Set later
        step.enabled = None
        step.will_exist = None
        step.otf_cache = True  # on-the-fly cache checking

    @property
    def node_id(step):
        """
        The experiment manager constructs output paths such that they
        are unique given the specific set of inputs and parameters. Thus
        the output paths are sufficient to determine a unique id per step.
        """
        return step.name + '_' + ub.hash_data(step.out_paths)[0:12]

    @property
    def command(step):
        if step.otf_cache and step.enabled != 'redo':
            return step.test_is_computed_command() + ' || ' + step._command
        else:
            return step._command

    def test_is_computed_command(step):
        test_expr = ' -a '.join(
            [f'-e "{p}"' for p in step.out_paths.values()])
        test_cmd = 'test ' +  test_expr
        return test_cmd

    @ub.memoize_property
    def does_exist(self):
        return all(self.out_paths.map_values(lambda p: p.exists()).values())


class Pipeline:
    """
    Registers how to call each step in the pipeline
    """

    @classmethod
    def connect_steps(cls, steps, skip_existing=False):
        """
        Build the graph that represents this pipeline
        """
        # Determine the interaction / dependencies between step inputs /
        # outputs
        g = nx.DiGraph()
        outputs_to_step = ub.ddict(list)
        inputs_to_step = ub.ddict(list)
        for step in steps.values():
            for path in step.out_paths.values():
                outputs_to_step[path].append(step.name)
            for path in step.in_paths.values():
                inputs_to_step[path].append(step.name)
            g.add_node(step.name, step=step)

        inputs_to_step = ub.udict(inputs_to_step)
        outputs_to_step = ub.udict(outputs_to_step)

        common = list((inputs_to_step & outputs_to_step).keys())
        for path in common:
            isteps = inputs_to_step[path]
            osteps = outputs_to_step[path]
            for istep, ostep in it.product(isteps, osteps):
                g.add_edge(ostep, istep)

        #
        # Determine which steps are enabled / disabled
        sorted_nodes = list(nx.topological_sort(g))
        g.graph['order'] = sorted_nodes

        for node in sorted_nodes:
            step = g.nodes[node]['step']
            # if config['skip_existing']:
            ancestors_will_exist = all(
                g.nodes[ancestor]['step'].will_exist
                for ancestor in nx.ancestors(g, step.name)
            )
            if skip_existing and step.enabled != 'redo' and step.does_exist:
                step.enabled = False
            step.will_exist = (
                (step.enabled and ancestors_will_exist) or
                step.does_exist
            )

        if 0:
            from cmd_queue.util.util_networkx import write_network_text
            write_network_text(g)
        return g

    def __init__(pipe):
        pass

    @staticmethod
    def _make_argstr(params):
        parts = [f'    --{k}={v} \\' for k, v in params.items()]
        return chr(10).join(parts).lstrip().rstrip('\\')

    def act_crop(crop_params, **paths):
        paths = ub.udict(paths)
        from watch.cli import coco_align
        confobj = coco_align.__config__
        known_args = set(confobj.default.keys())
        assert not len(ub.udict(crop_params) - known_args), 'unknown args'
        crop_params = {
            'geo_preprop': 'auto',
            'keep': 'img',
            'force_nodata': -9999,
            'rpc_align_method': 'orthorectify',
            'target_gsd': 4,
            'site_summary': True,
        } | ub.udict(crop_params)

        # The best setting of this depends on if the data is remote or not.
        # When networking, around 20+ workers is a good idea, but that's a very
        # bad idea for local images or if the images are too big.
        # Parametarizing would be best.
        CROP_IMAGE_WORKERS = 16
        CROP_AUX_WORKERS = 8

        perf_options = {
            'verbose': 1,
            'workers': CROP_IMAGE_WORKERS,
            'aux_workers': CROP_AUX_WORKERS,
            'debug_valid_regions': False,
            'visualize': False,
        }
        crop_kwargs = { **paths }
        crop_kwargs['crop_params_argstr'] = Pipeline._make_argstr(crop_params)
        crop_kwargs['crop_perf_argstr'] = Pipeline._make_argstr(perf_options)

        command = ub.codeblock(
            r'''
            python -m watch.cli.coco_align \
                --src "{crop_src_fpath}" \
                --dst "{crop_fpath}" \
                {crop_params_argstr} \
                {crop_perf_argstr} \
            ''').format(**crop_kwargs).strip().rstrip('\\')

        # FIXME: parametarize and only if we need secrets
        # secret_fpath = ub.Path('$HOME/code/watch/secrets/secrets').expand()
        # # if ub.Path.home().name.startswith('jon'):
        #     # if secret_fpath.exists():
        #     #     secret_fpath
        #         # command = f'source {secret_fpath} && ' + command
        command = 'AWS_DEFAULT_PROFILE=iarpa GDAL_DISABLE_READDIR_ON_OPEN=EMPTY_DIR ' + command

        name = 'crop'
        step = Step(name, command,
                    in_paths=paths & {'crop_regions', 'crop_src_fpath'},
                    out_paths=paths & {'crop_fpath', 'crop_dpath'},
                    resources={'cpus': 2})
        return step

    def pred_trk_pxl(perf_params, trk_pxl_params, **paths):
        paths = ub.udict(paths)
        workers = perf_params['trk.pxl.workers']
        perf_options = {
            'num_workers': workers,
            'devices': perf_params['trk.pxl.devices'],
            'accelerator': perf_params['trk.pxl.accelerator'],
            'batch_size': perf_params['trk.pxl.batch_size'],
        }
        paths = ub.udict(paths).map_values(lambda p: ub.Path(p).expand())
        pred_trk_pxl_kw =  { **paths }
        pred_trk_pxl_kw['params_argstr'] = Pipeline._make_argstr(trk_pxl_params)
        pred_trk_pxl_kw['perf_argstr'] = Pipeline._make_argstr(perf_options)
        command = ub.codeblock(
            r'''
            python -m watch.tasks.fusion.predict \
                --package_fpath={pkg_trk_pxl_fpath} \
                --test_dataset={trk_test_dataset_fpath} \
                --pred_dataset={pred_trk_pxl_fpath} \
                {params_argstr} \
                {perf_argstr}
            ''').format(**pred_trk_pxl_kw)
        name = 'pred_trk_pxl'
        step = Step(name, command,
                    in_paths=paths & {'pkg_trk_pxl_fpath',
                                      'trk_test_dataset_fpath'},
                    out_paths=paths & {'pred_trk_pxl_fpath'},
                    resources={'cpus': workers, 'gpus': 2})
        return step

    def pred_act_pxl(perf_params, act_pxl_params, **paths):
        """
        Idea:
            Step(<command_template>, <path-templates>)

            path_templates=[
                Template('pkg_act_pxl_fpath', 'packages/{act_expt}/{act_model}.pt')
                Template('act_test_dataset_fpath', '???'),
                Template('pred_act_pxl_fpath', '{pred_act_pxl_dpath}/pred.kwcoco.json'),
            ]
        """
        paths = ub.udict(paths)
        workers = perf_params['act.pxl.workers']
        perf_options = {
            'num_workers': workers,  # note: the "num_workers" cli arg should be changed to "workers" everywhere
            'devices': perf_params['act.pxl.devices'],
            'accelerator': perf_params['act.pxl.accelerator'],
            'batch_size': perf_params['act.pxl.batch_size'],
        }
        paths = ub.udict(paths).map_values(lambda p: ub.Path(p).expand())
        pred_act_pxl_kw =  { **paths }
        pred_act_pxl_kw['params_argstr'] = Pipeline._make_argstr(act_pxl_params)
        pred_act_pxl_kw['perf_argstr'] = Pipeline._make_argstr(perf_options)
        command = ub.codeblock(
            r'''
            python -m watch.tasks.fusion.predict \
                --package_fpath={pkg_act_pxl_fpath} \
                --test_dataset={act_test_dataset_fpath} \
                --pred_dataset={pred_act_pxl_fpath} \
                {params_argstr} \
                {perf_argstr}
            ''').format(**pred_act_pxl_kw)
        name = 'pred_act_pxl'
        step = Step(name, command,
                    in_paths=paths & {
                        'pkg_act_pxl_fpath',
                        'act_test_dataset_fpath'
                    },
                    out_paths=paths & {'pred_act_pxl_fpath'},
                    resources={'cpus': workers, 'gpus': 2})
        return step

    def pred_trk_poly(trk_poly_params, **paths):
        paths = ub.udict(paths)
        pred_trk_poly_kw = { **paths }
        cfg = trk_poly_params.copy()
        if 'thresh_hysteresis' in cfg:
            if isinstance(cfg['thresh_hysteresis'], str):
                cfg['thresh_hysteresis'] = util_globals.restricted_eval(
                    cfg['thresh_hysteresis'].format(**cfg))

        if 'moving_window_size' in cfg:
            if isinstance(cfg['moving_window_size'], str):
                cfg['moving_window_size'] = util_globals.restricted_eval(
                    cfg['moving_window_size'].format(**cfg))
        else:
            cfg['moving_window_size'] = None

        # kwargs['params_argstr'] = Pipeline(trk_poly_params)
        pred_trk_poly_kw['track_kwargs_str'] = shlex.quote(json.dumps(cfg))

        command = ub.codeblock(
            r'''
            python -m watch.cli.run_tracker \
                "{pred_trk_pxl_fpath}" \
                --default_track_fn saliency_heatmaps \
                --track_kwargs {track_kwargs_str} \
                --clear_annots \
                --out_sites_dir "{pred_trk_poly_sites_dpath}" \
                --out_site_summaries_dir "{pred_trk_poly_site_summaries_dpath}" \
                --out_sites_fpath "{pred_trk_poly_sites_fpath}" \
                --out_site_summaries_fpath "{pred_trk_poly_site_summaries_fpath}" \
                --out_kwcoco "{pred_trk_poly_kwcoco}"
            ''').format(**pred_trk_poly_kw)
        name = 'pred_trk_poly'
        step = Step(name, command,
                    in_paths=paths & {'pred_trk_pxl_fpath'},
                    out_paths=paths & {
                        'pred_trk_poly_dpath',
                        'pred_trk_poly_sites_fpath',
                        'pred_trk_poly_site_summaries_fpath',
                        'pred_trk_poly_kwcoco'},
                    resources={'cpus': 2})
        return step

    def pred_act_poly(act_poly_params, **paths):
        paths = ub.udict(paths)
        pred_act_poly_kw = paths.copy()
        # pred_act_row['site_summary_glob'] = (region_model_dpath / '*.geojson')
        # pred_act_poly_kw['site_summary_glob'] = (pred_act_poly_kw['region_model_dpath'] / '*.geojson')
        # pred_act_poly_kw['site_summary_glob'] = (paths['pred_trk_poly_sites_dpath'] / '*.geojson')
        pred_act_poly_kw['site_summary'] = act_poly_params.pop('site_summary')
        actclf_cfg = {
            'boundaries_as': 'polys',
        }
        actclf_cfg.update(act_poly_params)
        pred_act_poly_kw['kwargs_str'] = shlex.quote(json.dumps(actclf_cfg))
        command = ub.codeblock(
            r'''
            python -m watch.cli.run_tracker \
                "{pred_act_pxl_fpath}" \
                --default_track_fn class_heatmaps \
                --track_kwargs {kwargs_str} \
                --site_summary '{site_summary}' \
                --out_sites_fpath "{pred_act_poly_sites_fpath}" \
                --out_sites_dir "{pred_act_poly_sites_dpath}" \
                --out_kwcoco "{pred_act_poly_kwcoco}"
            ''').format(**pred_act_poly_kw)
        name = 'pred_act_poly'
        step = Step(name, command,
                    in_paths=paths & {'pred_act_pxl_fpath'},
                    out_paths=paths & {'pred_act_poly_dpath',
                                       'pred_act_poly_sites_fpath',
                                       'pred_act_poly_kwcoco'},
                    resources={'cpus': 2})
        return step

    def eval_trk_pxl(condensed, **paths):
        paths = ub.udict(paths)
        paths = ub.udict(paths).map_values(lambda p: ub.Path(p).expand())
        eval_trk_pxl_kw = { **paths }
        extra_opts = {
            'draw_curves': True,  # todo: parametarize
            'draw_heatmaps': True,  # todo: parametarize
            'viz_thresh': 0.2,
            'workers': 2,
        }
        eval_trk_pxl_kw['extra_argstr'] = Pipeline._make_argstr(extra_opts)
        command = ub.codeblock(
            r'''
            python -m watch.tasks.fusion.evaluate \
                --true_dataset={trk_test_dataset_fpath} \
                --pred_dataset={pred_trk_pxl_fpath} \
                --eval_dpath={eval_trk_pxl_dpath} \
                --score_space=video \
                {extra_argstr}
            ''').format(**eval_trk_pxl_kw)
        name = 'eval_trk_pxl'
        step = Step(name, command,
                    in_paths=paths & {'pred_trk_pxl_fpath'},
                    out_paths=paths & {'eval_trk_pxl_fpath',
                                       'eval_trk_pxl_dpath'},
                    resources={'cpus': 2})
        return step

    def eval_act_pxl(condensed, **paths):
        paths = ub.udict(paths)
        paths = ub.udict(paths).map_values(lambda p: ub.Path(p).expand())
        eval_act_pxl_kw = { **paths }
        extra_opts = {
            'draw_curves': True,
            'draw_heatmaps': True,
            'viz_thresh': 0.2,
            'workers': 2,
        }
        eval_act_pxl_kw['extra_argstr'] = Pipeline._make_argstr(extra_opts)
        command = ub.codeblock(
            r'''
            python -m watch.tasks.fusion.evaluate \
                --true_dataset={act_test_dataset_fpath} \
                --pred_dataset={pred_act_pxl_fpath} \
                --eval_dpath={eval_act_pxl_dpath} \
                --score_space=video \
                {extra_argstr}
            ''').format(**eval_act_pxl_kw).strip().rstrip('\\')
        name = 'eval_act_pxl'
        step = Step(name, command,
                    in_paths=paths & {'pred_act_pxl_fpath'},
                    out_paths=paths & {'eval_act_pxl_fpath',
                                       'eval_act_pxl_dpath'},
                    resources={'cpus': 2})
        return step

    def viz_pred_trk_poly(condensed, **paths):
        paths = ub.udict(paths)
        viz_pred_trk_poly_kw = paths.copy()
        viz_pred_trk_poly_kw['extra_header'] = f"\\n{condensed['trk_pxl_cfg']}-{condensed['trk_poly_cfg']}"
        viz_pred_trk_poly_kw['viz_channels'] = "red|green|blue,salient"
        command = ub.codeblock(
            r'''
            smartwatch visualize \
                "{pred_trk_poly_kwcoco}" \
                --channels="{viz_channels}" \
                --stack=only \
                --workers=avail/2 \
                --workers=avail/2 \
                --extra_header="{extra_header}" \
                --animate=True && touch {pred_trk_poly_viz_stamp}
            ''').format(**viz_pred_trk_poly_kw)
        name = 'viz_pred_trk_poly'
        step = Step(name, command,
                    in_paths=paths & {'pred_trk_poly_kwcoco'},
                    out_paths=paths & {'pred_trk_poly_viz_stamp'},
                    resources={'cpus': 2})
        return step

    def viz_pred_act_poly(condensed, **paths):
        paths = ub.udict(paths)
        viz_pred_act_poly_kw = paths.copy()
        viz_pred_act_poly_kw['extra_header'] = f"\\n{condensed['act_pxl_cfg']}-{condensed['act_poly_cfg']}"
        # viz_pred_act_poly_kw['viz_channels'] = 'red|green|blue,No Activity|Site Preparation|Active Construction|Post Construction'
        viz_pred_act_poly_kw['viz_channels'] = 'red|green|blue,No Activity|Active Construction|Post Construction'
        command = ub.codeblock(
            r'''
            smartwatch visualize \
                "{pred_act_poly_kwcoco}" \
                --channels="{viz_channels}" \
                --stack=only \
                --workers=avail/2 \
                --extra_header="{extra_header}" \
                --animate=True && touch {pred_act_poly_viz_stamp}
            ''').format(**viz_pred_act_poly_kw)
        name = 'viz_pred_act_poly'
        step = Step(name, command,
                    in_paths=paths & {'pred_act_poly_kwcoco'},
                    out_paths=paths & {'pred_act_poly_viz_stamp'},
                    resources={'cpus': 2})
        return step

    def eval_trk_poly(condensed, **paths):
        paths = ub.udict(paths)
        eval_trk_poly_kw = { **paths }
        eval_trk_poly_kw['eval_trk_poly_dpath'] = eval_trk_poly_kw['eval_trk_poly_dpath']
        eval_trk_poly_kw['eval_trk_poly_tmp_dpath'] = eval_trk_poly_kw['eval_trk_poly_dpath'] / '_tmp'
        eval_trk_poly_kw['name_suffix'] = '-'.join([
            condensed.get('trk_model', 'unk_trk_model'),
            condensed.get('trk_pxl_cfg', 'unk_trk_pxl_cfg'),
            condensed.get('trk_poly_cfg', 'unk_trk_poly_cfg'),
        ])
        command = ub.codeblock(
            r'''
            python -m watch.cli.run_metrics_framework \
                --merge=True \
                --name "{name_suffix}" \
                --true_site_dpath "{true_site_dpath}" \
                --true_region_dpath "{true_region_dpath}" \
                --pred_sites "{pred_trk_poly_sites_fpath}" \
                --tmp_dir "{eval_trk_poly_tmp_dpath}" \
                --out_dir "{eval_trk_poly_dpath}" \
                --merge_fpath "{eval_trk_poly_fpath}"
            ''').format(**eval_trk_poly_kw)
        name = 'eval_trk_poly'
        step = Step(name, command,
                    in_paths=paths & {'pred_trk_poly_sites_fpath'},
                    out_paths=paths & {'eval_trk_poly_dpath',
                                       'eval_trk_poly_fpath'},
                    resources={'cpus': 2})
        return step

    def eval_act_poly(condensed, **paths):
        """
        Idea:
            Step(<command_template>, <path-templates>)

            path_templates=[
                Template('true_site_dpath', 'true_site_dpath')
                Template('pred_act_poly_sites_fpath', '{pred_act_poly_dpath}/activity_tracks.kwcoco.json'),
            ]
        """
        paths = ub.udict(paths)
        eval_act_poly_kw = paths.copy()
        eval_act_poly_kw['name_suffix'] = '-'.join([
            # condensed['crop_dst_dset']
            condensed.get('test_act_dset', 'unk_act_dset'),
            condensed.get('act_model', 'unk_act_model'),
            condensed.get('act_pxl_cfg', 'unk_act_pxl_cfg'),
            condensed.get('act_poly_cfg', 'unk_act_poly_cfg'),
        ])
        eval_act_poly_kw['eval_act_poly_dpath'] = eval_act_poly_kw['eval_act_poly_dpath']
        eval_act_poly_kw['eval_act_poly_tmp_dpath'] = eval_act_poly_kw['eval_act_poly_dpath'] / '_tmp'
        command = ub.codeblock(
            r'''
            python -m watch.cli.run_metrics_framework \
                --merge=True \
                --name "{name_suffix}" \
                --true_site_dpath "{true_site_dpath}" \
                --true_region_dpath "{true_region_dpath}" \
                --pred_sites "{pred_act_poly_sites_fpath}" \
                --tmp_dir "{eval_act_poly_tmp_dpath}" \
                --out_dir "{eval_act_poly_dpath}" \
                --merge_fpath "{eval_act_poly_fpath}"
            ''').format(**eval_act_poly_kw)
        name = 'eval_act_poly'
        step = Step(name, command,
                    in_paths=paths & {'pred_act_poly_sites_fpath'},
                    out_paths=paths & {'eval_act_poly_fpath',
                                       'eval_act_poly_dpath'},
                    resources={'cpus': 2})
        return step


def submit_old_pipeline_jobs(resolved_rows, queue, config, annotations_dpath):

    # TODO: parameterize
    perf_params = {
        'trk.pxl.batch_size': 1,
        'trk.pxl.workers': 4,
        'trk.pxl.devices': '0,',
        'trk.pxl.accelerator': 'gpu',

        'act.pxl.batch_size': 1,
        'act.pxl.workers': 4,
        'act.pxl.devices': '0,',
        'act.pxl.accelerator': 'gpu',
    }

    common_submitkw = dict(
        partition=config['partition'],
        mem=config['mem']
    )

    # Each row represents a single source-to-sink pipeline run, but multiple
    # rows may share pipeline steps. This is handled by having unique ids per
    # job that depend on their outputs.
    for row in ub.ProgIter(resolved_rows, desc='submiting pipelines'):
        paths = row['paths']
        paths['true_annotations_dpath'] = annotations_dpath
        paths['true_site_dpath'] = paths['true_annotations_dpath'] / 'site_models'
        paths['true_region_dpath'] = paths['true_annotations_dpath'] / 'region_models'
        task_params = row['task_params']
        paths = ub.udict(paths).map_values(lambda p: ub.Path(p).expand())

        trk_pxl_params = task_params['trk.pxl']
        trk_poly_params = task_params['trk.poly']
        crop_params = task_params['crop']
        act_pxl_params = task_params['act.pxl']
        act_poly_params = task_params['act.poly']
        condensed = row['condensed']

        ### define the dag of this row item.

        steps = {}

        # TODO: detect if BAS team features are needed and compute those here.
        # Do we update the dataset paths here or beforehand?

        step = Pipeline.pred_trk_pxl(perf_params, trk_pxl_params, **paths)
        steps[step.name] = step

        step = Pipeline.pred_trk_poly(trk_poly_params, **paths)
        steps[step.name] = step

        step = Pipeline.act_crop(crop_params, **paths)
        steps[step.name] = step

        # TODO: detect if SC team features are needed and compute those here.

        step = Pipeline.pred_act_pxl(perf_params, act_pxl_params, **paths)
        steps[step.name] = step

        step = Pipeline.pred_act_poly(act_poly_params, **paths)
        steps[step.name] = step

        step = Pipeline.eval_trk_pxl(condensed, **paths)
        steps[step.name] = step

        step = Pipeline.eval_trk_poly(condensed, **paths)
        steps[step.name] = step

        step = Pipeline.viz_pred_trk_poly(condensed, **paths)
        steps[step.name] = step

        step = Pipeline.eval_act_pxl(condensed, **paths)
        steps[step.name] = step

        step = Pipeline.eval_act_poly(condensed, **paths)
        steps[step.name] = step

        step = Pipeline.viz_pred_act_poly(condensed, **paths)
        steps[step.name] = step

        for step in steps.values():
            step.enabled = config['enable_' + step.name]
            step.otf_cache = config['dynamic_skips']

        ADD_LINKS = config['enable_links']
        if ADD_LINKS:
            # Make jobs to symlink things on the fly
            linkable_pairs = [
                {'pred': 'pred_trk_pxl_dpath', 'eval': 'eval_trk_pxl_dpath'},
                {'pred': 'pred_act_pxl_dpath', 'eval': 'eval_act_pxl_dpath'},
                {'pred': 'pred_trk_poly_dpath', 'eval': 'eval_trk_poly_dpath'},
                {'pred': 'pred_act_poly_dpath', 'eval': 'eval_act_poly_dpath'},
            ]
            for pair in linkable_pairs:
                pred_key = pair['pred']
                eval_key = pair['eval']
                pred_dpath = paths[pred_key]
                eval_dpath = paths[eval_key]
                pred_lpath = eval_dpath / '_pred_link'
                eval_lpath = pred_dpath / '_eval_link'
                out_paths = {
                    pred_key.replace('dpath', 'lpath'): pred_lpath,
                    eval_key.replace('dpath', 'lpath'): eval_lpath,
                }
                name = 'link_' + eval_key.replace('eval_', '')
                parts = [
                    f'ln -sf "{eval_dpath}" "{eval_lpath}"',
                    f'ln -sf "{pred_dpath}" "{pred_lpath}"',
                ]
                command = ' && '.join(parts)
                step = Step(name, command,
                            in_paths=paths & {pred_key, eval_key},
                            out_paths=out_paths,
                            resources={'cpus': 1})
                steps[step.name] = step
                step.enabled = True
                step.otf_cache = False

            # The track to act is multiway.
            # One way pair
            if condensed['regions_id'] != 'truth':
                key1 = 'pred_trk_poly_dpath'
                key2 = 'crop_dpath'
                dpath1 = paths[key1]
                dpath2 = paths[key2]
                lpath_parent = dpath2 / '_crop_links'
                lpath1 = lpath_parent / condensed['crop_id']
                lpath2 = dpath1 / '_pred_trk_poly_link'
                eval_lpath = pred_dpath / '_eval_link'
                out_paths = {
                    key1.replace('dpath', 'lpath'): lpath1,
                    key2.replace('dpath', 'lpath'): lpath2,
                }
                name = 'link_trk_crop'
                parts = [
                    f'mkdir {lpath_parent}',
                    f'ln -sf "{dpath1}" "{lpath1}"',
                    f'ln -sf "{dpath2}" "{lpath2}"',
                ]
                command = ' && '.join(parts)
                step = Step(name, command,
                            in_paths=paths & {pred_key, eval_key},
                            out_paths=out_paths,
                            resources={'cpus': 1})
                steps[step.name] = step
                step.enabled = True
                step.otf_cache = True

        skip_existing = config['skip_existing']
        g = Pipeline.connect_steps(steps, skip_existing)

        #
        # Submit steps to the scheduling queue
        for node in g.graph['order']:
            # Skip duplicate jobs
            step = g.nodes[node]['step']
            if step.node_id in queue.named_jobs:
                continue
            depends = []
            for other, _ in list(g.in_edges(node)):
                dep_step = g.nodes[other]['step']
                if dep_step.enabled:
                    depends.append(dep_step.node_id)

            if step.will_exist and step.enabled:
                queue.submit(command=step.command, name=step.node_id,
                             depends=depends, **common_submitkw)



#!/usr/bin/env python3
"""
Attempts to register directory of geotiffs into a kwcoco dataset
"""
import scriptconfig as scfg
import ubelt as ub


class KWCocoFromGeotiffConfig(scfg.DataConfig):
    """
    Create a kwcoco manifest of a set of on-disk geotiffs
    """
    __default__ = {
        'geotiff_dpath': scfg.Value(None, help='path containing geotiffs'),
        'relative': scfg.Value(False, help='if true make paths relative'),
        'dst': scfg.Value(None, help='path to write new kwcoco file'),
        'workers': scfg.Value(0, help='number of parallel jobs'),
        'strict': scfg.Value(False, isflag=True, help='it True, will raise an error if any job fails'),
    }


def main(cmdline=1, **kwargs):
    """
    Ignore:
        geotiff_dpath = '/home/joncrall/data/grab_tiles_out/fels'
        dst = '/home/joncrall/data/grab_tiles_out/fels/data.kwcoco.json'
        kwargs = {
            'geotiff_dpath': geotiff_dpath,
            'dst': dst,
        }
        dset1 = kwcoco.CocoDataset(ub.expandpath('$HOME/data/dvc-repos/smart_watch_dvc/drop0/drop0.kwcoco.json'))
        dset2 = kwcoco.CocoDataset(ub.expandpath('$HOME/data/grab_tiles_out/fels/data.kwcoco.json'))
    """
    config = KWCocoFromGeotiffConfig.cli(data=kwargs, cmdline=cmdline)
    geotiff_dpath = config['geotiff_dpath']
    dst = config['dst']

    imgs = find_geotiffs(geotiff_dpath, workers=config['workers'],
                         strict=config['strict'])

    import kwcoco
    dset = kwcoco.CocoDataset()
    for img in imgs:
        dset.add_image(**img)
    return dset
    dset.fpath = dst

    if config['relative']:
        dset.reroot(dset.bundle_dpath, absolute=False)

    dset.dump(dset.fpath, newlines=True)


def filter_band_files(fpaths, band_list, with_tci=True):
    """
    band_list is any subset of util_bands.ALL_BANDS

    with_tci: include true color thumbnail
    """
    from os.path import basename, splitext
    band_names = set(b['name'] for b in band_list)
    if with_tci:
        band_names.add('TCI')
    # use endswith() instead of in
    # to avoid false positives, eg from a tile code in the filename

    def is_band_file(path):
        return any(splitext(basename(path))[0].endswith(b) for b in band_names)
    return list(filter(is_band_file, fpaths))


def ingest_landsat_directory(lc_dpath):
    import watch
    from watch.utils import util_bands
    from dateutil.parser import isoparse
    from os.path import join, basename, normpath
    import glob
    name = basename(normpath(lc_dpath))
    tiffs = sorted(glob.glob(join(lc_dpath, '*.TIF')))
    if len(tiffs) == 0:
        tiffs = sorted(glob.glob(join(lc_dpath, '**', '*.TIF'), recursive=True))
    baseinfo = watch.gis.geotiff.geotiff_filepath_info(name)
    capture_time = isoparse(baseinfo['filename_meta']['acquisition_date']).isoformat()
    sensor_coarse = 'LS'
    if baseinfo['filename_meta']['sensor_code'] == 'C':
        sensor_coarse = 'L8'
    elif baseinfo['filename_meta']['sensor_code'] == 'E':
        sensor_coarse = 'L7'
    # take L8 as the default guess for a mangled name
    tiffs = filter_band_files(tiffs, (util_bands.LANDSAT7 if sensor_coarse == 'L7' else util_bands.LANDSAT8))
    img = make_coco_img_from_auxiliary_geotiffs(tiffs, name)
    img['date_captured'] = capture_time
    img['sensor_coarse'] = sensor_coarse
    return img


def ingest_sentinel2_directory(s2_dpath):
    # Are we in the safedir, the granuledir or some arbitrary dir?
    # Try to use the granuledir as name if available;
    # it's a better unique ID.
    import watch
    from watch.utils import util_bands
    import datetime as datetime_mod
    from dateutil.parser import isoparse
    from os.path import join, basename, normpath
    import glob
    granules = sorted(glob.glob(join(s2_dpath, 'GRANULE', '*')))
    if len(granules) == 1:
        granule = granules[0]
        tiffs = sorted(glob.glob(join(granule, 'IMG_DATA', '*.jp2')))
        name = basename(normpath(granule))
    else:
        tiffs = sorted(glob.glob(join(s2_dpath, 'GRANULE', '*', 'IMG_DATA', '*.jp2')))
        if len(tiffs) == 0:
            tiffs = sorted(glob.glob(join(s2_dpath, '**', '*.jp2'), recursive=True))
        name = basename(normpath(s2_dpath)).replace('.SAFE', '')
    # Then grab the bands.
    tiffs = filter_band_files(tiffs, util_bands.SENTINEL2)
    img = make_coco_img_from_auxiliary_geotiffs(tiffs, name)

    baseinfo = watch.gis.geotiff.geotiff_filepath_info(s2_dpath)
    capture_time = isoparse(baseinfo['filename_meta']['sense_start_time'])
    img['date_captured'] = datetime_mod.datetime.isoformat(capture_time)
    img['sensor_coarse'] = 'S2'
    return img


def make_coco_img_from_geotiff(tiff_fpath, name=None, force_affine=True,
                               with_info=False):
    """
    TODO: move to coco extensions

    Example:
        >>> from watch.demo.landsat_demodata import grab_landsat_product  # NOQA
        >>> product = grab_landsat_product()
        >>> tiffs = product['bands'] + [product['meta']['bqa']]
        >>> tiff_fpath = product['bands'][0]
        >>> name = None
        >>> img = make_coco_img_from_geotiff(tiff_fpath)
        >>> print('img = {}'.format(ub.urepr(img, nl=1)))
    """
    import watch
    from watch.utils import util_bands
    import kwimage
    img = {}
    if name is not None:
        img['name'] = name

    info = watch.gis.geotiff.geotiff_metadata(tiff_fpath)
    # only affine transformations are supported in auxiliary channels
    # TODO support RPC
    info.update(**watch.gis.geotiff.geotiff_crs_info(tiff_fpath, force_affine=force_affine))

    warp_pxl_from_wld = kwimage.Affine.coerce(info['pxl_to_wld'])
    height, width = info['img_shape']
    file_meta = info['filename_meta']
    channels = file_meta.get('channels', None)

    if channels is None:
        # fix this for known WV channel signature, which isn't obvious from filename
        if file_meta.get('product_guess') == 'worldview':

            from osgeo import gdal
            bands = gdal.Info(tiff_fpath, format='json')['bands']

            # the channel names are the same for all WV, just the center_wavelength is different
            # so we can safely use this info from WV2
            def _code(band_dicts):
                return '|'.join(b.get('common_name', b['name']) for b in band_dicts)

            if len(bands) == 1:
                channels = _code(util_bands.WORLDVIEW2_PAN)
            elif len(bands) == 4:
                channels = _code(util_bands.WORLDVIEW2_MS4)
            elif len(bands) == 8:
                channels = _code(util_bands.WORLDVIEW2_MS8)
            else:
                raise Exception('unknown channel signature for WV')
        else:
            raise Exception('must be able to introspect channels')

    wld_crs_info = ub.dict_diff(info['wld_crs_info'], {'type'})
    utm_crs_info = ub.dict_diff(info['utm_crs_info'], {'type'})

    img.update({
        'file_name': tiff_fpath,
        'width': width,
        'height': height,
        'channels': channels,
        'num_bands': info['num_bands'],
        'approx_meter_gsd': info['approx_meter_gsd'],
        'warp_pxl_to_wld': warp_pxl_from_wld,
        'geos_corners': info['geos_corners'],
        'wld_crs_info': wld_crs_info,
        'utm_crs_info': utm_crs_info,
    })
    if with_info:
        img['info'] = info
    return img


def make_coco_img_from_auxiliary_geotiffs(tiffs, name):
    """
    TODO: move to coco extensions

    Example:
        >>> from watch.demo.landsat_demodata import grab_landsat_product  # NOQA
        >>> product = grab_landsat_product()
        >>> tiffs = product['bands'] + [product['meta']['bqa']]
        >>> name = product['scene_name']
        >>> img = make_coco_img_from_auxiliary_geotiffs(tiffs, name)
        >>> print('img = {}'.format(ub.urepr(img, nl=-1, sort=0)))
    """
    auxiliary = []
    for fpath in tiffs:
        aux = make_coco_img_from_geotiff(fpath, force_affine=True)
        auxiliary.append(aux)
    return make_coco_img_from_auxiliary_dicts(auxiliary, name)


def make_coco_img_from_auxiliary_dicts(auxiliary, name):
    img = {
        'name': name,
        'file_name': None,
    }
    # Choose a base image canvas and the relationship between auxiliary images
    idx = ub.argmax(auxiliary, lambda x: (x['width'] * x['height']))
    base = auxiliary[idx]
    warp_wld_from_img = base['warp_pxl_to_wld']
    warp_img_from_wld = warp_wld_from_img.inv()
    img['warp_img_to_wld'] = warp_wld_from_img.concise()
    img.update(ub.dict_isect(base, {'geos_corners', 'wld_crs_info', 'utm_crs_info'}))

    for aux in auxiliary:
        aux.pop('geos_corners')
        aux.pop('utm_crs_info')
        aux.pop('wld_crs_info')
        warp_wld_from_aux = aux.pop('warp_pxl_to_wld')
        warp_img_from_aux = warp_img_from_wld @ warp_wld_from_aux
        aux['warp_aux_to_img'] = warp_img_from_aux.concise()

    img['width'] = base['width']
    img['height'] = base['height']
    img['auxiliary'] = auxiliary
    return img


def find_geotiffs(geotiff_dpath, workers=0, strict=False):
    """
    Search a directory for any geotiffs and return a set of kwcoco-style image
    dictionaries detailing the results.

    Args:
        geotiff_dpath (str): directory to search

    Returns:
        List[Dict]: a list of kwcoco-style image dictionaries

    geotiff_dpath = '/home/joncrall/data/grab_tiles_out/fels'
    """
    import os
    import watch
    from os.path import basename
    dpath_list = list(watch.gis.geotiff.walk_geotiff_products(geotiff_dpath))

    print(f'Found candidate {len(dpath_list)} geotiff products')

    jobs = ub.JobPool(mode='thread', max_workers=workers)

    loose_files = []
    unknown_products = []
    for dpath in ub.ProgIter(dpath_list, desc='submit geotiffs jobs'):
        if os.path.isfile(dpath):
            # if we dont't get a directory we have a loose geotiff file
            loose_files.append(dpath)
        else:
            dname = basename(dpath)
            if dname.startswith(('LC', 'LE07')):
                lc_dpath = dpath
                job = jobs.submit(ingest_landsat_directory, lc_dpath)
                job.dpath = dpath
            elif dname.startswith('S2'):
                s2_dpath = dpath
                # FIXME: undefined name
                job = jobs.submit(ingest_sentinel2_directory, s2_dpath)
                job.dpath = dpath
            else:
                msg = ('unknown dpath = {!r}'.format(dpath))
                print(msg)
                unknown_products.append(msg)

    if unknown_products:
        print('\n'.join(unknown_products))

    imgs = []
    errors = []
    for job in ub.ProgIter(jobs, desc='collect results'):
        try:
            img = job.result()
        except Exception as ex:
            if strict:
                raise
            err = (job, job.dpath, ex)
            print('err = {!r}'.format(err))
            errors.append(err)
        else:
            imgs.append(img)

    if loose_files:
        # Handle loose files (try grouping them by spacetime)
        groups = ub.ddict(list)

        # jobs = ub.JobPool(mode='thread', max_workers=workers)
        for fpath in ub.ProgIter(loose_files, desc='process loose files'):
            info = watch.gis.geotiff.geotiff_filepath_info(fpath, fast=True)
            file_meta = info['filename_meta']
            file_meta.get('tile_number', None)
            date_captured = next(iter(ub.dict_isect(file_meta, ['sense_start_time', 'acquisition_date']).values()), None)
            tile_num = next(iter(ub.dict_isect(file_meta, ['tile_number']).values()), None)
            groupid = (file_meta['product_guess'], tile_num, date_captured)
            img = make_coco_img_from_geotiff(fpath, with_info=True)
            info = img.pop('info')
            img['date_captured'] = info['filename_meta']['date_captured']
            if 'landsat' in info['filename_meta']['product_guess']:
                sensor_coarse = 'LS'
                if info['filename_meta']['sensor_code'] == 'C':
                    sensor_coarse = 'L8'
                elif info['filename_meta']['sensor_code'] == 'E':
                    sensor_coarse = 'L7'
            elif 'sentinel' in info['filename_meta']['product_guess']:
                sensor_coarse = 'S2'
            else:
                sensor_coarse = 'null'
            img['sensor_coarse'] = sensor_coarse
            groups[groupid].append(img)

        for groupid, group in groups.items():
            img = make_coco_img_from_auxiliary_dicts(group, name=groupid)
            img['date_captured'] = group[0]['date_captured']
            img['sensor_coarse'] = group[0]['sensor_coarse']
            imgs.append(img)

    print('Got {} errors'.format(len(errors)))
    print('Found {} images to add'.format(len(imgs)))
    return imgs


__config__ = KWCocoFromGeotiffConfig


if __name__ == '__main__':
    """
    CommandLine:
        python -m watch.cli.geotiffs_to_kwcoco.py

    CommandLine:
        python -m watch.cli.coco_extract_geo_bounds \
          --src $HOME/data/dvc-repos/smart_watch_dvc/drop0/drop0.kwcoco.json \
          --breakup_times=True \
          --dst $HOME/data/grab_tiles_out/regions.geojson.json

        source ~/internal/safe/secrets
        python ~/code/watch/scripts/grab_tiles_demo.py \
            --regions $HOME/data/grab_tiles_out/regions.geojson.json \
            --out_dpath $HOME/data/grab_tiles_out \
            --rgdc_username=$WATCH_RGD_USERNAME \
            --rgdc_password=$WATCH_RGD_PASSWORD \
            --backend rgdc

        python ~/code/watch/scripts/grab_tiles_demo.py \
            --regions $HOME/data/grab_tiles_out/regions.geojson.json \
            --out_dpath $HOME/data/grab_tiles_out \
            --backend fels --profile

        python -m watch.cli.geotiffs_to_kwcoco.py \
            --geotiff_dpath ~/data/grab_tiles_out/fels \
            --dst $HOME/data/grab_tiles_out/fels/data.kwcoco.json --profile

        python -m watch.cli.geotiffs_to_kwcoco.py \
            --geotiff_dpath ~/data/dvc-repos/smart_watch_dvc/unannotated/AE-Dubai-0001 \
            --dst ~/data/dvc-repos/smart_watch_dvc/unannotated/dubai-msi.kwcoco.json

        cat ~/data/dvc-repos/smart_watch_dvc/unannotated/dubai-msi.kwcoco.json

        python -m watch.cli.geotiffs_to_kwcoco.py \
            --geotiff_dpath ~/data/dvc-repos/smart_watch_dvc/unannotated/KR-Pyeongchang-S2 \
            --dst ~/data/dvc-repos/smart_watch_dvc/unannotated/korea-msi.kwcoco.json

        python -m -m watch.cli.geotiffs_to_kwcoco.py \
            --geotiff_dpath ~/data/dvc-repos/smart_watch_dvc/unannotated/US-Waynesboro-0001 \
            --dst ~/data/dvc-repos/smart_watch_dvc/unannotated/waynesboro-msi.kwcoco.json
    """
    main()



# flake8: noqa

from PIL import Image
import torchvision.transforms.functional as FT
import torchvision
import torch
import numpy as np
from scipy.spatial import distance
import random
# import torchvision.transforms.functional.pad
# torch.backends.cudnn.enabled = True
# torch.backends.cudnn.deterministic = True
import itertools
from watch.tasks.rutgers_material_seg.utils import utils

# if 1:
#     torch.set_printoptions(precision=6, sci_mode=False)

IMG_EXTENSIONS = ['*.png', '*.jpeg', '*.jpg', '*.npy']


# FIXME: Hard coding mean/std is not a good idea
mean_std = ([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])


# class RandomCrop(object):

#     def __init__(self, size, padding=None, pad_if_needed=True, fill=0, padding_mode='constant'):

#         self.shape = size
#         self.padding = padding
#         self.pad_if_needed = pad_if_needed
#         self.fill = fill
#         self.padding_mode = padding_mode

#     @staticmethod
#     def get_params(img, output_size):

#         w, h, c  = img.shape
#         th, tw = output_size
#         if w == tw and h == th:
#             return 0, 0, h, w

#         i = random.randint(0, h - th)
#         j = random.randint(0, w - tw)
#         return i, j, th, tw

#     def __call__(self, image, mask):

#         i, j, h, w = self.get_params(image, self.shape)
#         crop_image = FT.crop(image, i, j, h, w)
#         crop_mask = FT.crop(mask, i, j, h, w)

#         return crop_image, crop_mask


class S2MCPDataset(object):
    def __init__(self, root, transforms, split=False, crop_size=300):
        self.root = root
        self.transforms = transforms
        self.split = split
        self.crop_size = crop_size
        self.randomcrop_transform = torchvision.transforms.RandomCrop(size=(crop_size, crop_size))

        self.images_root = f"{self.root}/"
        self.images1_paths = utils.dictionary_contents(path=self.images_root, types=['*a.npy'])

    def __getitem__(self, idx):
        negative_idx = random.choice([i for i in range(0, self.__len__()) if i != idx])

        img1_path = self.images1_paths[idx]
        image_name = img1_path.split('/')[-1].split('.')[0].split('_')[0]

        img2_path = f"{self.images_root}/{image_name}_b.npy"

        negative_image_name = self.images1_paths[negative_idx].split('/')[-1].split('.')[0]
        negative_image_path = f"{self.images_root}/{negative_image_name}.npy"

        img1 = np.load(img1_path)[:, :, :13]
        img2 = np.load(img2_path)[:, :, :13]
        negative_img = np.load(negative_image_path)

        new_image1 = self.transforms(img1)
        new_image2 = self.transforms(img2)
        new_negative_image = self.transforms(negative_img)

        crop_params = self.randomcrop_transform.get_params(new_image1, output_size=(self.crop_size, self.crop_size))
        new_image1 = FT.crop(new_image1, *crop_params)
        new_image2 = FT.crop(new_image2, *crop_params)
        new_negative_image = FT.crop(new_negative_image, *crop_params)

        outputs = {}
        outputs['visuals'] = {'image1': new_image1, 'image2': new_image2, 'image_name': image_name}
        outputs['inputs'] = {'image1': new_image1, 'image2': new_image2, 'negative_image': new_negative_image}

        return outputs

    def __len__(self):
        return len(self.images1_paths)



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


class ConfusorAnalysisConfig(scfg.DataConfig):
    """
    Requires that IARPA metrics are computed
    """
    detections_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        truth assignments
        usually detections_tau=0.2_rho=0.5_min_area=0.csv
        '''))
    proposals_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        detection assignments
        usually proposals_tau=0.2_rho=0.5_min_area=0.csv
        '''))

    src_kwcoco = scfg.Value(None, help='the input kwcoco file to project onto')
    dst_kwcoco = scfg.Value(None, help='the reprojected output kwcoco file to write')

    pred_sites = scfg.Value(None, help='the path to the predicted sites manifest / directory / globstr')
    bas_metric_dpath = scfg.Value(None, help='A path to bas metrics if det/prop paths are not specified')

    region_id = scfg.Value(None, help='the id for the region')
    true_site_dpath = scfg.Value(None, help='input')
    true_region_dpath = scfg.Value(None, help='input')

    performer_id = scfg.Value('kit', help='the performer id')

    out_dpath = scfg.Value(None, help='where to write results')

    def __post_init__(self):
        if self.bas_metric_dpath is not None:
            self.bas_metric_dpath = ub.Path(self.bas_metric_dpath)

        if self.true_region_dpath is not None:
            self.true_region_dpath = ub.Path(self.true_region_dpath)

        if self.true_site_dpath is not None:
            self.true_site_dpath = ub.Path(self.true_site_dpath)

        if self.out_dpath is not None:
            self.out_dpath = ub.Path(self.out_dpath)


def main(cmdline=1, **kwargs):
    """
    CommandLine:
        xdoctest -m /home/joncrall/code/watch/watch/mlops/confusor_analysis.py main
        HAS_DVC=1 xdoctest -m watch.mlops.confusor_analysis main:0

    Example:
        >>> # xdoctest: +REQUIRES(env:HAS_DVC)
        >>> from watch.mlops.confusor_analysis import *  # NOQA
        >>> import watch
        >>> data_dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> region_id = 'NZ_R001'
        >>> true_site_dpath = data_dvc_dpath / 'annotations/drop6/site_models'
        >>> true_region_dpath = data_dvc_dpath / 'annotations/drop6/region_models'
        >>> src_kwcoco = data_dvc_dpath / f'Drop6-MeanYear10GSD/imgonly-{region_id}.kwcoco.zip'
        >>> dst_kwcoco = data_dvc_dpath / f'Drop6-MeanYear10GSD/confusor-{region_id}.kwcoco.zip'
        >>> dag_dpath = ub.Path('/data/joncrall/dvc-repos/smart_expt_dvc/_airflow/ta2_preeval10_pyenv_t33_post3')
        >>> dpath = dag_dpath / region_id
        >>> bas_metric_dpath = dpath / 'metrics/overall/bas'
        >>> #bas_metric_dpath = dpath / 'local_metrics' / region_id / 'overall/bas'
        >>> out_dpath = dpath / 'local_metrics'
        >>> cmdline = 0
        >>> kwargs = dict(
        >>>     bas_metric_dpath=bas_metric_dpath,
        >>>     pred_sites=(dpath / 'sc-fusion/sc_out_site_models'),
        >>>     true_site_dpath=true_site_dpath,
        >>>     true_region_dpath=true_region_dpath,
        >>>     out_dpath=out_dpath,
        >>>     dst_kwcoco=dst_kwcoco,
        >>>     src_kwcoco=src_kwcoco,
        >>>     region_id=region_id,
        >>> )
        >>> main(cmdline=cmdline, **kwargs)
    """
    config = ConfusorAnalysisConfig.cli(cmdline=cmdline, data=kwargs, strict=True)

    import rich
    import pandas as pd
    import itertools as it
    from watch import heuristics
    from watch.utils import util_gis
    rich.print('config = ' + ub.urepr(config, nl=1, align=':'))
    region_id = config.region_id

    performer_id = config.performer_id
    true_site_dpath = config.true_site_dpath
    true_region_dpath = config.true_region_dpath
    pred_site_fpaths = list(util_gis.coerce_geojson_paths(config.pred_sites))

    if config.bas_metric_dpath is not None:
        if config.detections_fpath is None:
            config.detections_fpath = config.bas_metric_dpath / 'detections_tau=0.2_rho=0.5_min_area=0.csv'
        if config.proposals_fpath is None:
            config.proposals_fpath = config.bas_metric_dpath / 'proposals_tau=0.2_rho=0.5_min_area=0.csv'

    config.detections_fpath = ub.Path(config.detections_fpath)
    config.proposals_fpath = ub.Path(config.proposals_fpath)
    assert config.proposals_fpath.exists()
    assert config.detections_fpath.exists()

    assign1 = pd.read_csv(config.detections_fpath)
    assign2 = pd.read_csv(config.proposals_fpath)

    needs_recompute = any('_seq_' in m or m.startswith('seq_') for m in assign2['site model'] if m)
    assert not needs_recompute

    def fix_site_id(site_id):
        # Hack because idk why the metrics code does this.
        if site_id.startswith('_'):
            site_id = region_id + site_id
        return site_id

    ### Assign a confusion label to each truth and predicted annotation
    true_confusion_rows = []
    pred_confusion_rows = []
    site_to_status = {}
    for row in assign1.to_dict('records'):
        true_site_id = row['truth site'].split('_te_')[0]
        true_site_id = fix_site_id(true_site_id)
        pred_site_ids = []
        truth_status = row['site type']
        site_to_status[true_site_id] = truth_status
        if isinstance(row['matched site models'], str):
            for name in row['matched site models'].split(','):
                pred_site_id = name.strip().split(f'_{performer_id}_')[0]
                pred_site_id = fix_site_id(pred_site_id)
                pred_site_ids.append(pred_site_id)
        has_positive_match = len(pred_site_ids)
        true_cfsn = heuristics.iarpa_assign_truth_confusion(truth_status, has_positive_match)
        true_confusion_rows.append({
            'true_site_id': true_site_id,
            'pred_site_ids': pred_site_ids,
            'type': true_cfsn,
        })

    for row in assign2.to_dict('records'):
        pred_site_id = row['site model'].split(f'_{performer_id}_')[0]
        pred_site_id = fix_site_id(pred_site_id)
        true_site_ids = []
        truth_match_statuses = []
        if isinstance(row['matched truth sites'], str):
            for name in row['matched truth sites'].split(','):
                true_site_id = name.strip().split('_te_')[0]
                true_site_id = fix_site_id(true_site_id)
                truth_match_statuses.append(site_to_status[true_site_id])
                true_site_ids.append(true_site_id)
        pred_cfsn = heuristics.iarpa_assign_pred_confusion(truth_match_statuses)
        pred_confusion_rows.append({
            'pred_site_id': pred_site_id,
            'true_site_ids': true_site_ids,
            'type': pred_cfsn,
        })

    for row in true_confusion_rows + pred_confusion_rows:
        row['color'] = heuristics.IARPA_CONFUSION_COLORS.get(row['type'])

    """
    True Confusion Spec
    -------------------

    "misc_info":  {
        "confusion": {
            "true_site_id": str,          # redundant site id information,
            "pred_site_ids": List[str],   # the matching predicted site ids,
            "type": str,                  # the type of true confusion assigned by T&E
            "color": str,                 # a named color coercable via kwimage.Color.coerce
        }
    }

    Predicted Confusion Spec
    -------------------

    "misc_info":  {
        "confusion": {
            "pred_site_id": str,          # redundant site id information,
            "true_site_ids": List[str],   # the matching predicted site ids,
            "type": str,        # the type of predicted confusion assigned by T&E
            "color": str,       # a named color coercable via kwimage.Color.coerce
        }
    }

    # The possible confusion codes and the corresponding confusion_color they
    # will be assigned is defined in heuristics.IARPA_CONFUSION_COLORS
    """

    # Add the confusion info as misc data in new site files and reproject them
    # onto the truth for visualization.
    # pred_sites_fpath = poly_pred_dpath / 'sites_manifest.json'
    # assert pred_sites_fpath.exists()
    # pred_site_fpaths = list(util_gis.coerce_geojson_paths(pred_sites_fpath))

    rm_files = list(true_region_dpath.glob(region_id + '*.geojson'))

    gt_files = list(true_site_dpath.glob(region_id + '*.geojson'))
    sm_files = pred_site_fpaths
    true_site_infos = list(util_gis.coerce_geojson_datas(gt_files, format='json'))
    pred_site_infos = list(util_gis.coerce_geojson_datas(sm_files, format='json'))
    orig_region_infos = list(util_gis.coerce_geojson_datas(rm_files, format='json'))
    assert len(orig_region_infos) == 1

    # Ensure all site data has misc-info
    # Ensure all data cast to site models
    from watch.geoannots.geomodels import SiteModel, SiteModelCollection, RegionModel
    import json

    true_region_model = RegionModel(orig_region_infos[0]['data'])

    for info in it.chain(pred_site_infos, true_site_infos):
        info['data'] = SiteModel(**info['data'])
        info['data'].header['properties'].setdefault('misc_info', {})

    id_to_true_data = {ub.Path(d['fpath']).stem: d for d in true_site_infos}
    id_to_pred_data = {ub.Path(d['fpath']).stem: d for d in pred_site_infos}

    # Add confusion metadata to predicted and truth models
    # https://gis.stackexchange.com/questions/346518/opening-geojson-style-properties-in-qgis
    for row in true_confusion_rows:
        info = id_to_true_data[row['true_site_id']]
        info['data'].header['properties']['misc_info']['confusion'] = row
    for row in pred_confusion_rows:
        info = id_to_pred_data[row['pred_site_id']]
        info['data'].header['properties']['misc_info']['confusion'] = row

    pred_sites = SiteModelCollection([d['data'] for d in pred_site_infos])
    true_sites = SiteModelCollection([d['data'] for d in true_site_infos])

    VALIDATE = 1
    if VALIDATE:
        all_models = SiteModelCollection(pred_sites + true_sites)
        all_models.fixup()
        all_models.validate(workers=8)

    pred_region_model = pred_sites.as_region_model(region=true_region_model.header)
    pred_df = pred_region_model.pandas_summaries()
    true_df = true_region_model.pandas_summaries()
    idx1_to_idx2 = util_gis.geopandas_pairwise_overlaps(pred_df, true_df)

    # Find predicted annotations that have no truth overlap
    non_intersecting_idxs = [idx1 for idx1, idx2s in idx1_to_idx2.items() if not len(idx2s)]
    cand_df = pred_df.iloc[non_intersecting_idxs]

    hard_positive_site_ids = []
    for true_site in true_sites:
        misc = true_site.header['properties']['misc_info']
        if misc['confusion']['type'] in 'gt_false_neg':
            hard_positive_site_ids.append(true_site.header['properties']['site_id'])

    hard_negative_sites = []
    for site_id in cand_df['site_id']:
        pred_site = id_to_pred_data[site_id]['data']
        misc = pred_site.header['properties']['misc_info']
        if misc['confusion']['type'] in 'sm_completely_wrong':
            hard_negative_sites.append(pred_site.deepcopy())

    orig_site_num = int(true_df['site_id'].max().split('_')[-1])
    base_site_num = max(900, orig_site_num)
    for num, hard_neg in enumerate(hard_negative_sites, start=base_site_num):
        header_prop = hard_neg.header['properties']
        header_prop['site_id'] = region_id + f'_{num:04d}'
        header_prop['status'] = 'negative'
        header_prop['model_content'] = 'annotation'
        header_prop['misc_info'].pop('confusion', None)
        header_prop['misc_info']['kwcoco'] = {'weight': 1.5}
        header_prop['comments'] = 'hard_negative'
        for obs in hard_neg.observations():
            props = obs['properties']
            props['current_phase'] = None
            props['is_site_boundary'] = None
            props.pop('predicted_phase_transition', None)
            props.pop('predicted_phase_transition_date', None)

    hard_neg_summaries = [s.as_summary() for s in hard_negative_sites]
    new_region = true_region_model.deepcopy()
    new_region['features'] += hard_neg_summaries
    new_true_sites = [s.deepcopy() for s in true_sites]

    # Upweight hard positive true sites
    print('hard_positive_site_ids = {}'.format(ub.urepr(hard_positive_site_ids, nl=1)))
    new_true_props = [n.header['properties'] for n in new_true_sites] + [s['properties'] for s in new_region.site_summaries()]
    for prop in new_true_props:
        if 'misc_info' not in prop:
            prop['misc_info'] = {}
        prop['misc_info'].pop('confusion', None)
        if prop['site_id'] in hard_positive_site_ids:
            prop['misc_info']['kwcoco'] = {'weight': 2.0}
            if 'comments' not in prop or not prop['comments']:
                prop['comments'] = 'hard_positive'
            else:
                prop['comments'] += ';hard_positive'

    # Add in hard negatives
    new_sites = new_true_sites + hard_negative_sites

    enriched_dpath = config.out_dpath / 'enriched_annots'
    enriched_sites_dpath = enriched_dpath / 'site_models'
    enriched_region_dpath = enriched_dpath / 'region_models'

    enriched_sites_dpath.ensuredir()
    enriched_region_dpath.ensuredir()

    rich.print(f'enriched_dpath: [link={enriched_dpath}]{enriched_dpath}[/link]')
    new_region_fpath = enriched_region_dpath / (region_id + '.geojson')
    new_region_fpath.write_text(json.dumps(new_region, indent='    '))
    for new_site in new_sites:
        fpath = enriched_sites_dpath / (new_site.site_id + '.geojson')
        text = json.dumps(new_site, indent='    ')
        fpath.write_text(text)

    # Dump confusion categorized site models to disk
    cfsn_dpath = config.out_dpath / 'confusion_sites'
    true_cfsn_dpath = (cfsn_dpath / 'true').ensuredir()
    pred_cfsn_dpath = (cfsn_dpath / 'pred').ensuredir()
    for pred_site_id, pred_site in id_to_pred_data.items():
        fpath = pred_cfsn_dpath / (pred_site_id + '.geojson')
        text = json.dumps(pred_site['data'], indent='    ')
        fpath.write_text(text)
    for true_site_id, true_site in id_to_true_data.items():
        fpath = true_cfsn_dpath / (true_site_id + '.geojson')
        text = json.dumps(true_site['data'], indent='    ')
        fpath.write_text(text)

    # Need to build site summaries from site models.

    # Write a new "enriched truth" file that reweights false negatives add
    # false positive as hard negatives.

    REPROJECT = config.src_kwcoco is not None
    if REPROJECT:
        from watch.cli import reproject_annotations
        new_coco_fpath = enriched_dpath / f'hardneg-{region_id}.kwcoco.zip'
        common_kwargs = ub.udict(
            src=config.src_kwcoco,
            dst=new_coco_fpath,
            site_models=enriched_sites_dpath,
            region_models=enriched_region_dpath,
            workers=2,
        )
        reproject_annotations.main(cmdline=0, **common_kwargs)

        if 0:
            # Project confusion site models onto kwcoco for visualization
            import kwcoco
            from watch.cli import reproject_annotations
            src_dset = kwcoco.CocoDataset(config.src_kwcoco)
            dst_dset = src_dset.copy()
            dst_dset.fpath = config.dst_kwcoco
            dst_dset.clear_annotations()
            true_site_infos2 = list(util_gis.coerce_geojson_datas(
                id_to_true_data.values(), format='dataframe', allow_raw=True))
            pred_site_infos2 = list(util_gis.coerce_geojson_datas(
                id_to_pred_data.values(), format='dataframe', allow_raw=True))

            for info in pred_site_infos2:
                site_df = info['data']

            for info in true_site_infos2:
                site_df = info['data']
                reproject_annotations.validate_site_dataframe(site_df)

            dst_dset.clear_annotations()
            common_kwargs = ub.udict(
                clear_existing=False,
                src=dst_dset,
                dst='return',
                workers=2,
            )
            true_kwargs = common_kwargs | ub.udict(
                role='true_confusion',
                # propogate_strategy=False,
                # propogate_strategy=False,
                site_models=true_site_infos2,
                # viz_dpath=(out_dpath / '_true_projection'),
            )
            kwargs = true_kwargs
            pred_kwargs = common_kwargs | ub.udict(
                role='pred_confusion',
                site_models=pred_site_infos2,
                # viz_dpath=(out_dpath / '_pred_projection'),
            )
            # I don't know why this isn't in-place. Maybe it is a scriptconfig thing?
            repr1 = str(dst_dset.annots())
            print(f'repr1={repr1}')
            dst_dset = reproject_annotations.main(cmdline=0, **true_kwargs)
            repr2 = str(dst_dset.annots())
            print(f'repr1={repr1}')
            print(f'repr2={repr2}')
            pred_kwargs['src'] = dst_dset
            dst_dset = reproject_annotations.main(cmdline=0, **pred_kwargs)
            repr3 = str(dst_dset.annots())
            print(f'repr1={repr1}')
            print(f'repr2={repr2}')
            print(f'repr3={repr3}')

            set(dst_dset.annots().lookup('role', None))
            set([x['role'] for x in dst_dset.annots().lookup('misc_info', None)])
            # dst_dset.annots().take([0, 1, 2])
            viz_dpath = cfsn_dpath
            summary_visualization(dst_dset, viz_dpath)
            rich.print(f'Viz Dpath: [link={viz_dpath}]{viz_dpath}[/link]')


def summary_visualization(dst_dset, viz_dpath):
    import kwplot
    import numpy as np

    resolution = '5GSD'

    from watch.utils import util_progress
    # from watch.utils import util_kwimage
    import kwarray
    import kwimage

    pman = util_progress.ProgressManager()
    with pman:
        # video_to_tracking_heatmap = {}
        for video in pman(dst_dset.videos().objs, desc='make video summary'):
            # Simulate the tracking heatmap (todo: get what the data really was)
            images = dst_dset.images(video_id=video['id'])
            det_accum = []
            for coco_img in pman(images.coco_images, desc='load dets'):
                dets = coco_img._detections_for_resolution(resolution=resolution)
                det_accum.append(dets)

            running = kwarray.RunningStats()
            jobs = ub.JobPool('process', max_workers=2, transient=True)

            chan_priority = [
                'red|green|blue',
                'pan',
            ]

            def acceptable_channels(channels, chan_priority):
                import kwcoco
                for want in chan_priority:
                    want = kwcoco.FusedChannelSpec.coerce(want)
                    cand = (channels & want)
                    if cand.numel() == want.numel():
                        return cand
                return None

            if 1:
                for coco_img in pman(images.coco_images, desc='submit delay jobs'):
                    chanels = acceptable_channels(coco_img.channels, chan_priority)
                    delayed = coco_img.imdelay(chanels, resolution=resolution, nodata_method='float')
                    job = jobs.submit(delayed.finalize)

                    from watch.utils import util_time
                    time = util_time.coerce_datetime(coco_img['timestamp'])
                    job.time = time
                    job.coco_img = coco_img

                oldest_time = None
                oldest_img = None
                newest_time = None
                newest_img = None

                job_loader = pman(jobs.as_completed(), total=len(jobs), desc='averaging heatmaps')
                for job in job_loader:
                    im = job.result()

                    if oldest_time is None or oldest_time > job.time:
                        if job.coco_img['sensor_coarse'] == 'S2':
                            oldest_time = job.time
                            oldest_img = im

                    if newest_time is None or newest_time < job.time:
                        if job.coco_img['sensor_coarse'] == 'S2':
                            newest_time = job.time
                            newest_img = im

                    im = kwimage.atleast_3channels(im)
                    running.update(im)
                    # track_ids = dst_dset.annots(dets.data['aids']).lookup('track_id')

            all_dets = kwimage.Detections.concatenate(det_accum)
            all_annots = dst_dset.annots(all_dets.data['aids'])
            all_dets.data['frame_index'] = np.array(all_annots.images.lookup('frame_index'))
            all_dets.data['track_id'] = np.array(all_annots.lookup('track_id'))
            all_dets.data['role'] = np.array(all_annots.lookup('role'))
            all_dets.data['misc_info'] = np.array(all_annots.lookup('misc_info'))

            groupers = list(zip(all_dets.data['role'], all_dets.data['track_id']))
            unique_tids, groupxs = kwarray.group_indices(groupers)

            track_summaries = []
            from shapely.ops import unary_union
            for (role, tid), groupx in zip(unique_tids, groupxs):
                track_dets = all_dets.take(groupx)
                misc_info = track_dets.data['misc_info'][0]
                row = misc_info.copy()
                # row['role'] = role
                # assert row['role'] == role
                sh_poly = unary_union([p.to_shapely() for p in track_dets.data['segmentations']])
                kw_poly = kwimage.MultiPolygon.from_shapely(sh_poly)
                row['poly'] = kw_poly
                track_summaries.append(row)

            # canvas = kwplot.make_heatmask(util_kwimage.exactly_1channel(mean_heatmap), cmap='magma')[:, :, 0:3]

            old_canvas = kwimage.normalize_intensity(oldest_img, axis=2)
            new_canvas = kwimage.normalize_intensity(newest_img, axis=2)

            current = running.current()
            mean_heatmap = current['mean']
            # min_heatmap = current['min']
            # max_heatmap = current['max']
            canvas = kwimage.normalize_intensity(mean_heatmap)
            # canvas = kwplot.make_heatmask(util_kwimage.exactly_1channel(mean_heatmap), cmap='viridis')[:, :, 0:3]
            # canvas_raw = canvas.copy()
            # canvas_true = canvas.copy()
            # canvas_pred = canvas.copy()
            # canvas_cfsn = canvas.copy()

            canvases = {
                'new': {
                    'true': new_canvas.copy(),
                    'pred': new_canvas.copy(),
                    'cfsn': new_canvas.copy(),
                },
                'old': {
                    'true': old_canvas.copy(),
                    'pred': old_canvas.copy(),
                    'cfsn': old_canvas.copy(),
                },
                'avg': {
                    'true': canvas.copy(),
                    'pred': canvas.copy(),
                    'cfsn': canvas.copy(),
                },
            }

            role_to_summary = ub.udict(ub.group_items(track_summaries, key=lambda x: x['role']))
            print(ub.udict(role_to_summary).map_values(len))

            alpha = 0.6

            for row in ub.ProgIter(role_to_summary['true_confusion']):
                for k1, v1 in canvases.items():
                    v1['true'] = row['poly'].draw_on(v1['true'], fill=False, edgecolor=row['confusion_color'], alpha=alpha)
                    v1['cfsn'] = row['poly'].draw_on(v1['cfsn'], fill=False, edgecolor=row['confusion_color'], alpha=alpha)
                # row['poly'].draw_on(canvas_true, fill=False, edgecolor=row['confusion_color'])
                # row['poly'].draw_on(canvas_cfsn, fill=False, edgecolor=row['confusion_color'])

            for row in ub.ProgIter(role_to_summary['pred_confusion']):
                for k1, v1 in canvases.items():
                    v1['pred'] = row['poly'].draw_on(v1['pred'], fill=False, edgecolor=row['confusion_color'], alpha=alpha)
                    v1['cfsn'] = row['poly'].draw_on(v1['cfsn'], fill=False, edgecolor=row['confusion_color'], alpha=alpha)
                # row['poly'].draw_on(canvas_pred, fill=False, edgecolor=row['confusion_color'])
                # row['poly'].draw_on(canvas_cfsn, fill=False, edgecolor=row['confusion_color'])

            for k1, v1 in canvases.items():
                v1['true'] = kwimage.draw_header_text(v1['true'], 'true confusion')
                v1['pred'] = kwimage.draw_header_text(v1['pred'], 'pred confusion')
                v1['cfsn'] = kwimage.draw_header_text(v1['cfsn'], 'both')

            row1 = kwimage.stack_images(list(canvases['avg'].values()), axis=1, pad=10)
            row2 = kwimage.stack_images(list(canvases['old'].values()), axis=1, pad=10)
            row3 = kwimage.stack_images(list(canvases['new'].values()), axis=1, pad=10)
            row1 = kwimage.ensure_uint255(row1)
            row2 = kwimage.ensure_uint255(row2)
            row3 = kwimage.ensure_uint255(row3)

            final_canvas = kwimage.stack_images([row1, row2, row3], axis=0, pad=5)

            fpath = viz_dpath / f'confusion_{video["name"]}.jpg'
            kwimage.imwrite(fpath, final_canvas)

    from watch import heuristics
    legend_img = kwplot.make_legend_img(heuristics.IARPA_CONFUSION_COLORS)
    kwimage.imwrite(viz_dpath / 'confusion_legend.png', legend_img)


if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/watch/watch/mlops/confusor_analysis.py
        python -m watch.mlops.confusor_analysis
    """
    main()



def fix_msys_path(path):
    r"""
    Windows is so special. When using msys bash if you pass a path on the CLI
    it resolves /c to C:/, but if you have you path as part of a config string,
    it doesnt know how to do that, and at that point Python doesn't handle the
    msys style /c paths. This is a hack detects and fixes this in this
    location.

    Example:
        >>> from watch.utils.util_windows import *  # NOQA
        >>> print(fix_msys_path('/c/Users/foobar'))
        C:/Users/foobar
        >>> print(fix_msys_path(r'\c\Users\foobar'))
        C:/Users\foobar
        >>> print(fix_msys_path(r'\d\Users\foobar'))
        D:/Users\foobar
        >>> print(fix_msys_path(r'\z'))
        Z:/
        >>> import pathlib
        >>> assert fix_msys_path(pathlib.Path(r'\z')) == pathlib.Path('Z:/')
    """
    import os
    was_pathlike = isinstance(path, os.PathLike)
    input_path = path
    new_path = path
    if was_pathlike:
        path = os.fspath(path)
    if isinstance(path, str):
        import re
        drive_pat = '(?P<drive>[A-Za-z])'
        slash_pat = r'[/\\]'
        pat1 = re.compile(f'^{slash_pat}{drive_pat}$')
        pat2 = re.compile(f'^{slash_pat}{drive_pat}{slash_pat}.*$')
        match = pat1.match(path) or pat2.match(path)
        if match is not None:
            drive_name = match.groupdict()['drive'].upper()
            new_path = f'{drive_name}:/{path[3:]}'
    if was_pathlike:
        new_path = input_path.__class__(new_path)
    return new_path


def is_windows_path(path):
    r"""
    Example:
        >>> from watch.utils.util_windows import *  # NOQA
        >>> assert is_windows_path('C:')
        >>> assert is_windows_path('C:/')
        >>> assert is_windows_path('C:\\')
        >>> assert is_windows_path('C:/foo')
        >>> assert is_windows_path('C:\\foo')
        >>> assert not is_windows_path('/foo')
    """
    import re
    drive_pat = '(?P<drive>[A-Za-z])'
    slash_pat = r'[/\\]'
    pat1 = re.compile(f'^{drive_pat}:$')
    pat2 = re.compile(f'^{drive_pat}:{slash_pat}.*$')
    match = pat1.match(path) or pat2.match(path)
    return bool(match)



from torch.utils.data import Dataset
import torch
import numpy as np
import os
import albumentations as A
import kwcoco
import kwimage
import random
import ubelt as ub
import tifffile as tif


class kwcoco_dataset(Dataset):
    S2_l2a_channel_names = [
        'B02.tif', 'B01.tif', 'B03.tif', 'B04.tif', 'B05.tif', 'B06.tif', 'B07.tif', 'B08.tif', 'B09.tif', 'B11.tif', 'B12.tif', 'B8A.tif'
    ]
    S2_channel_names = [
        'coastal', 'blue', 'green', 'red', 'B05', 'B06', 'B07', 'nir', 'B09', 'cirrus', 'swir16', 'swir22', 'B8A'
    ]
    L8_channel_names = [
        'coastal', 'lwir11', 'lwir12', 'blue', 'green', 'red', 'nir', 'swir16', 'swir22', 'pan', 'cirrus'
    ]

    def __init__(self, coco_dset, sensor=['S2', 'L8'], bands=['shared'], patch_size=64, segmentation_labels=False, display=False, num_images=3, mode='train'):
        # initialize dataset
        self.dset = kwcoco.CocoDataset.coerce(coco_dset)
        self.images = self.dset.images()
        self.segmentation_labels = segmentation_labels
        self.annotations = self.dset.annots
        self.display = display

        self.num_images = num_images

        if not ub.iterable(sensor):
            sensor = [sensor]
        if not ub.iterable(bands):
            bands = [sensor]

        requested_sensors = set(sensor)

        if type(sensor) is not list:
            sensor = [sensor]
        if type(bands) is not list:
            bands = [bands]

        # handle if there are multiple sensors
        image_sensors = self.images.lookup('sensor_coarse', default=None)
        avail_sensors = set(image_sensors)
        flags = [x in requested_sensors for x in image_sensors]
        self.images = self.images.compress(flags)

        # if 'sensor_coarse' in self.images._id_to_obj[self.images._ids[0]].keys():
        #     # get available sensors
        #     # filter images by desired sensor
        #     self.images = self.images.compress([x in sensor for x in self.images.lookup('sensor_coarse')])
        #     assert(self.images)
        # else:
        #     avail_sensors = None
        print('avail_sensors:', avail_sensors)
        print('requested_sensors:', requested_sensors)
        self.dset_ids = self.images.gids
        self.videos = [x['id'] for x in self.dset.videos().objs]

        # get all available channels
        all_channels = [ aux.get('channels', None) for aux in self.dset.index.imgs[self.images._ids[0]].get('auxiliary', []) ]
        if 'r|g|b' in all_channels:
            all_channels.remove('r|g|b')
        self.channels = []
        # no channels selected
        if len(bands) < 1:
            raise ValueError(f'bands must be specified. Options are {", ".join(all_channels)}, or all')
        # all channels selected
        elif len(bands) == 1:

            if bands[0].lower() == 'all':
                self.channels = all_channels
            elif bands[0].lower() == 'shared':
                self.channels = ['red', 'green', 'blue', 'nir', 'swir16', 'swir22']
            elif bands[0] == 'r|g|b':
                self.channels.append('r|g|b')
            # subset of channels selected
        else:
            for band in bands:
                if band in all_channels:
                    self.channels.append(band)
                else:
                    raise ValueError(f'\'{band}\' not recognized as an available band. Options are {", ".join(all_channels)}, or all')

        # define augmentations
        additional_targets = dict()
        for i in range(self.num_images):
            additional_targets['image{}'.format(1 + i)] = 'image'
            additional_targets['seg{}'.format(i + 1)] = 'mask'

        if mode == 'test':
            self.segmentation_labels = False
            self.transforms = A.Compose([
                A.NoOp(),
                ],
                additional_targets=additional_targets)
            self.transforms2 = A.Compose([
                    A.VerticalFlip(p=.999),
                    A.HorizontalFlip(p=.999)],
                additional_targets={'image2': 'image'})

            self.transforms3 = A.Compose([
                    A.Blur(p=.75),
                    A.RandomBrightnessContrast(brightness_by_max=False, always_apply=True)
            ])
        else:
            self.transforms = A.Compose([
                A.RandomCrop(height=patch_size, width=patch_size),
                A.RandomRotate90(p=.999)
                ],
                additional_targets=additional_targets)

        self.num_channels = len(self.channels)
        self.bands = bands

        self.mode = mode

    def __len__(self,):
        return len(self.dset_ids)

    def get_img(self, idx, device=None):
        img1_id = self.dset_ids[idx]
        video = [y for y in self.videos if img1_id in self.dset.index.vidid_to_gids[y]][0]

        # randomly select image2 id from the same video (could be before or after image1)
        # make sure image2 is not image1 and image2 is in the set of filtered images by desired sensor
        id_list = [img1_id]
        frame_index = {1: self.dset.index.imgs[img1_id]['frame_index']}
        sensor_list = [self.dset.index.imgs[img1_id]['sensor_coarse']]
        for j in range(1, self.num_images):
            new_img_id = img1_id
            while (new_img_id in id_list) or (new_img_id not in self.dset_ids):
                new_img_id = random.choice(self.dset.index.vidid_to_gids[video])
            id_list.append(new_img_id)
            frame_index[1 + j] = self.dset.index.imgs[new_img_id]['frame_index']
            sensor_list.append(self.dset.index.imgs[new_img_id]['sensor_coarse'])
        # load images
        image_dict = dict()
        for k in range(self.num_images):
            image = self.dset.delayed_load(id_list[k], channels=self.channels, space='video').finalize().astype(np.float32)

            image = np.nan_to_num(image)
            if image.std() != 0.:
                image = (image - image.mean()) / image.std()
            image_dict['img{}'.format(1 + k)] = image

        image = [torch.tensor(image_dict[key]).permute(2, 0, 1).unsqueeze(0) for key in image_dict]

        image = torch.stack(image, dim=1)
        image_info = self.dset.index.imgs[img1_id]

        image = torch.nan_to_num(image)

        if device:
            image = image.to(device)
        # normalize
        if image.std() != 0.0:
            image = (image - image.mean()) / image.std()

        return img1_id, image_info, image

    def __getitem__(self, idx):
        # get image1 id and the video it is associated with
        img1_id = self.dset_ids[idx]
        video = [y for y in self.videos if img1_id in self.dset.index.vidid_to_gids[y]][0]
        out = dict()
        # randomly select image2 id from the same video (could be before or after image1)
        # make sure image2 is not image1 and image2 is in the set of filtered images by desired sensor
        id_list = [img1_id]
        img1_info = self.dset.index.imgs[img1_id]
        frame_index = {1: img1_info['frame_index']}
        sensor_list = [img1_info['sensor_coarse']]
        date = img1_info['date_captured']
        date_list = [torch.tensor([int(date[:4]), int(date[5:7])])]
        for j in range(1, self.num_images):
            new_img_id = img1_id
            while (new_img_id in id_list) or (new_img_id not in self.dset_ids):
                new_img_id = random.choice(self.dset.index.vidid_to_gids[video])
            id_list.append(new_img_id)
            info = self.dset.index.imgs[new_img_id]
            frame_index[1 + j] = info['frame_index']
            sensor_list.append(info['sensor_coarse'])
            date = info['date_captured']
            date_list.append(torch.tensor([int(date[:4]), int(date[5:7])]))
        # load images
        image_dict = dict()
        for k in range(self.num_images):
            image = self.dset.delayed_load(id_list[k], channels=self.channels, space='video').finalize().astype(np.float32)

            image = np.nan_to_num(image)
            if image.std() != 0.:
                image = (image - image.mean()) / image.std()
            image_dict['img{}'.format(1 + k)] = image

        albumentations_input = {'image': image_dict['img1']}
        for k in range(1, self.num_images):
            albumentations_input['image{}'.format(1 + k)] = image_dict['img{}'.format(1 + k)]

        if not self.segmentation_labels:
            # transformations
            transformed = self.transforms(**albumentations_input)
            images = [transformed[key] for key in transformed]

            if self.mode == 'test':
                # additional transforms for test mode
                transformed = self.transforms(image=images[0], image2=images[1])
                transformed2 = self.transforms2(image=images[0])
                img1 = transformed['image']
                img2 = transformed['image2']
                img3 = transformed2['image']
                transformed3 = self.transforms3(image=img1)
                # convert to tensors
                img1 = torch.tensor(img1).permute(2, 0, 1)
                img2 = torch.tensor(img2).permute(2, 0, 1)
                img4 = transformed3['image']
                img3 = torch.tensor(img3).permute(2, 0, 1)
                img4 = torch.tensor(img4).permute(2, 0, 1)

                frame_index1 = self.dset.index.imgs[img1_id]['frame_index']
                frame_index2 = self.dset.index.imgs[id_list[1]]['frame_index']

                out['offset_image1'] = img3.float()
                out['augmented_image1'] = img4.float()
                out['img1_id'] = img1_id
                out['time_sort_label'] = float(frame_index1 < frame_index2)
                # out['img1_info'] = self.dset.index.imgs[img1_id]

            images = [torch.tensor(x).permute(2, 0, 1) for x in images]

            if self.display:
                if self.num_channels == 3:

                    display_image1 = images[0]
                    display_image2 = images[1]
                elif self.num_channels < 10:
                    display_image1 = images[0][[0, 1, 2], :, :]
                    display_image2 = images[1][[0, 1, 2], :, :]
                else:
                    display_image1 = images[0][[3, 2, 1], :, :]
                    display_image2 = images[1][[3, 2, 1], :, :]
                display_image1 = (2 + display_image1) / 6
                display_image2 = (2 + display_image2) / 6
            else:
                display_image1 = torch.tensor([])
                display_image2 = torch.tensor([])

            for m in range(self.num_images):
                out['image{}'.format(1 + m)] = images[m]
                out['frame_number{}'.format(1 + m)] = frame_index[1 + m]

            out['display_image1'] = display_image1
            out['display_image2'] = display_image2

        else:
            aids = [self.dset.index.gid_to_aids[n] for n in id_list]

            dets = [kwimage.Detections.from_coco_annots(self.dset.annots(p).objs, dset=self.dset) for p in aids]

            vid_from_img = [kwimage.Affine.coerce(self.dset.index.imgs[q]['warp_img_to_vid']) for q in id_list]

            dets = [det.warp(img) for (det, img) in zip(dets, vid_from_img)]

            # bbox = dets.data['boxes'].data
            segmentations = []
            category_id = []
            for r in range(self.num_images):
                segmentations.append(dets[r].data['segmentations'].data)
                category_id.append([dets[r].classes.idx_to_id[cidx] for cidx in dets[r].data['class_idxs']])

            img_dims = (image_dict['img1'].shape[0], image_dict['img1'].shape[1])

            masks = dict()

            for s in range(self.num_images):
                combined = []
                for sseg, cid in zip(segmentations[s], category_id[s]):
                    assert cid > 0
                    np_mask = sseg.to_mask(dims=img_dims).data.astype(float) * cid
                    mask = torch.from_numpy(np_mask)
                    combined.append(mask.unsqueeze(0))
                if combined:
                    overall_mask = torch.max(torch.cat(combined, dim=0), dim=0)[0].numpy()  # ### HACK
                else:
                    overall_mask = np.zeros_like(image_dict['img1'][:, :, 0])

                masks['seg{}'.format(1 + s)] = overall_mask

            albumentations_input = {'image': image_dict['img1'], 'mask': masks['seg1']}
            for t in range(1, self.num_images):
                albumentations_input['image{}'.format(1 + t)] = image_dict['img{}'.format(1 + t)]
                albumentations_input['seg{}'.format(1 + t)] = masks['seg{}'.format(1 + t)]

            transformed = self.transforms(**albumentations_input)
            images = [transformed[key] for key in transformed if 'image' in key]

            images = [torch.tensor(x).permute(2, 0, 1) for x in images]
            all_segmentations = [transformed[key] for key in transformed if 'seg' in key or 'mask' in key]
            all_segmentations = [torch.tensor(x) for x in all_segmentations]
            segmentations = [torch.where(seg == 2, 1, 0) + torch.where(seg == 3, 1, 0) + torch.where(seg == 8, 1, 0) - torch.where(seg == 5, 1, 0) - torch.where(seg == 6, 1, 0) for seg in all_segmentations]

            if self.display:
                if self.num_channels == 3:
                    display_image1 = images[0]
                    display_image2 = images[1]
                elif self.num_channels < 10:
                    display_image1 = images[0][[0, 1, 2], :, :]
                    display_image2 = images[1][[0, 1, 2], :, :]
                else:
                    display_image1 = images[0][[3, 2, 1], :, :]
                    display_image2 = images[1][[3, 2, 1], :, :]
                display_image1 = (2 + display_image1) / 6
                display_image2 = (2 + display_image2) / 6
            else:
                display_image1 = torch.tensor([])
                display_image2 = torch.tensor([])

            for m in range(self.num_images):
                out['image{}'.format(1 + m)] = images[m]
                out['frame_number{}'.format(1 + m)] = frame_index[1 + m]
                out['segmentation{}'.format(1 + m)] = segmentations[m]
            out['display_image1'] = display_image1
            out['display_image2'] = display_image2
            out['order'] = [x[0] for x in sorted(frame_index.items(), key=lambda item: item[1])]

        out['frame_number'] = torch.tensor([out[key] for key in out if 'frame_number' in key])
        out['normalized_date'] = torch.tensor([date_[0] - 2018 + date_[1] / 12 for date_ in date_list])
        out['sensors'] = sensor_list
        # out['img1_info'] = img1_info
        out['img1_id'] = img1_id
        return out


class SpaceNet7(Dataset):
    '''TO DO: Return segmentation labels for SpaceNet 7'''
    normalize_params = [[0.16198677, 0.22665408, 0.1745371], [0.06108317, 0.06515977, 0.04128775]]

    def __init__(self,
                    patch_size=128,
                    splits='satellite_sort/data/spacenet/splits_unmasked/',  # ### unmasked images
                    train=True,
                    display=False,
                    num_images=3,
                    segmentation_labels=False):
        self.num_images = num_images
        self.display = display
        self.train = train
        self.segmentation_labels = segmentation_labels

        additional_targets = dict()
        for i in range(1, self.num_images):
            additional_targets['image{}'.format(1 + i)] = 'image'
            additional_targets['mask{}'.format(1 + i)] = 'mask'

        self.transforms = A.Compose([
                #                 A.NoOp(),
                A.RandomCrop(height=patch_size, width=patch_size),
                A.RandomRotate90(p=.999)
                ],
                additional_targets=additional_targets)

        if train:
            self.images = []
            with open(os.path.join(splits, 'spacenet7_train_images.txt'), 'r') as file:
                for row in file:
                    self.images.append(row.strip('\n'))

            self.segmentation_masks = []
            if segmentation_labels:
                with open(os.path.join(splits, 'spacenet7_train_building_masks.txt'), 'r') as file:
                    for row in file:
                        self.segmentation_masks.append(row.strip('\n'))
        else:
            self.images = []
            with open(os.path.join(splits, 'spacenet7_val_images.txt'), 'r') as file:
                for row in file:
                    self.images.append(row.strip('\n'))

            self.segmentation_masks = []
            if segmentation_labels:
                with open(os.path.join(splits, 'spacenet7_val_building_masks.txt'), 'r') as file:
                    for row in file:
                        self.segmentation_masks.append(row.strip('\n'))

        self.length = len(self.images)

    def date_to_step(self, date):
        year, month = date
        return (year % 2018) * 12 + month

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        images = dict()

        im1_path = self.images[idx]
        date1 = (int(im1_path[-47:-43]), int(im1_path[-42:-40]))

        im_paths = [im1_path]
        dates = [date1]

        im_directory, _ = os.path.split(self.images[idx])
        sorted_img_directory = sorted(os.listdir(im_directory))

        date1 = (int(self.images[idx][-47:-43]), int(self.images[idx][-42:-40]))
        image1 = tif.imread(im1_path).astype('float32')
        images['img1'] = image1

        if self.segmentation_labels:
            segmentations = dict()
            mask_directory, _ = os.path.split(self.segmentation_masks[idx])
            sorted_mask_directory = sorted(os.listdir(mask_directory))
            segmentations['seg1'] = tif.imread(self.segmentation_masks[idx]).astype('float32') / 255

        for x in range(1, self.num_images):
            date_new = date1
            while date_new in dates:
                idx_new = torch.randint(len(sorted_img_directory), (1,))[0]

                im_new_path = os.path.join(im_directory, sorted_img_directory[idx_new])
                date_new = (int(im_new_path[-47:-43]), int(im_new_path[-42:-40]))
            if self.segmentation_labels:
                segmentations_new_path = os.path.join(mask_directory, sorted_mask_directory[idx_new])
                segmentations['seg{}'.format(1 + x)] = tif.imread(segmentations_new_path).astype('float32') / 255
            dates.append(date_new)
            im_paths.append(im_new_path)
            image_new = tif.imread(im_new_path).astype('float32')
            images['img{}'.format(1 + x)] = image_new

        albumentations_input = {'image': images['img1']}
        if self.segmentation_labels:
            albumentations_input['mask'] = segmentations['seg1']
        for t in range(1, self.num_images):
            albumentations_input['image{}'.format(1 + t)] = images['img{}'.format(1 + t)]
            if self.segmentation_labels:
                albumentations_input['mask{}'.format(1 + t)] = segmentations['seg{}'.format(1 + t)]

        transformed = self.transforms(**albumentations_input)

        images = [transformed[key][:, :, :3] for key in transformed if 'image' in key]

        normalized_images = [torch.tensor((x - x.mean()) / x.std()).permute(2, 0, 1) for x in images[:self.num_images]]

        cloud_masks = [transformed[key][:, :, 3] / 255 for key in transformed if 'image' in key]

        if self.display:
            display_images = [x.astype('float32') for x in images]
        else:
            display_images = [torch.tensor([]) for x in range(self.num_images)]
        if self.segmentation_labels:
            segmentations = [torch.tensor(transformed[key]) for key in transformed if 'mask' in key]

        out = dict()
        for m in range(self.num_images):
            out['image{}'.format(1 + m)] = normalized_images[m]
            out['date{}'.format(1 + m)] = self.date_to_step(dates[m])
            out['cloud_mask{}'.format(1 + m)] = cloud_masks[m]
            if self.segmentation_labels:
                out['segmentation{}'.format(1 + m)] = segmentations[m]
            if self.display:
                out['display_image{}'.format(1 + m)] = display_images[m]

        sort_order = torch.tensor([y[0] + y[1] / 100 for y in dates[:self.num_images]])
        out['order'] = sort_order.sort()[1]
        out['time_steps'] = torch.tensor([out[key] for key in out if 'date' in key])

        return out



"""
Developer notebook
"""


def dzyne_mwe():
    import watch
    import ubelt as ub
    from watch.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDataset
    dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
    coco_fpath = dvc_dpath / 'Drop4-SC/yourdata.kwcoco.json'
    channels = 'red|green|blue,change'
    self = KWCocoVideoDataset(coco_fpath,
                              time_steps=5,
                              chip_dims=(196, 196),
                              time_sampling='uniform',
                              input_resolution='3GSD',
                              window_resolution='3GSD',
                              output_resolution='3GSD',
                              channels=channels)
    self.disable_augmenter = True

    # Get a sample target around a positive example.
    # Change properties as necessary to reproduce the error.
    target = self.new_sample_grid['targets'][self.new_sample_grid['positives_indexes'][0]].copy()

    # Execute sampling logic. Dive into the getitem function as necessary to
    # debug.
    item = self[target]
    print('item summary: ' + ub.urepr(self.summarize_item(item), nl=3))

    # Display the sample
    canvas = self.draw_item(item, overlay_on_image=0, rescale=1,
                            max_channels=5)

    import kwplot
    kwplot.autompl()
    kwplot.imshow(canvas, fnum=1)
    kwplot.show_if_requested()


def visualize_cloudmask_batch():
    # import os
    # os.environ['XDEV_PROFILE'] = '1'
    import watch
    import numpy as np  # NOQA
    import kwimage  # NOQA
    from watch.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDataset
    dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
    coco_fpath = dvc_dpath / 'Drop4-BAS/combo_vali_I2.kwcoco.json'
    # channels = 'red|green|blue,invariants.0:3,invariants.16,cloudmask'
    channels = 'red|green|blue,cloudmask'
    # coco_fpath = dvc_dpath / 'Drop4-BAS/KR_R001.kwcoco.json'
    # channels = 'red|green|blue,nir|swir16|swir22'
    self = KWCocoVideoDataset(coco_fpath,
                              time_steps=4,
                              chip_dims='full',
                              time_sampling='uniform',
                              input_resolution='native',
                              # input_resolution='3.3GSD',
                              window_resolution='10GSD',
                              output_resolution='10GSD',
                              channels=channels)
    self.disable_augmenter = True
    target = self.new_sample_grid['targets'][self.new_sample_grid['positives_indexes'][100]].copy()

    # vidid = self.sampler.dset.videos()[0]
    # images = self.sampler.dset.images(video_id=vidid)
    # full_gids = []
    # partial_gids = []
    # for obj in images.objs:
    #     poly = kwimage.MultiPolygon.coerce(obj['valid_region'])
    #     covered = poly.area / (obj['width'] * obj['height'])
    #     if covered >= 0.95:
    #         full_gids.append(obj['id'])
    #     else:
    #         partial_gids.append(obj['id'])
    # target['gids'] = full_gids[0:5] + partial_gids[0:2]
    # target['space_slice'] = (slice(-16, 224), slice(-16, 360))

    target['SAMECOLOR_QUALITY_HEURISTIC'] = None
    # target['SAMECOLOR_QUALITY_HEURISTIC'] = 'histogram'
    target['PROPOGATE_NAN_BANDS'] = set({'red'})
    # target['PROPOGATE_NAN_BANDS'] = set()
    target['force_bad_frames'] = 1
    target['mask_low_quality'] = 1
    target['quality_threshold'] = 0.0
    target['observable_threshold'] = 0.0
    target['resample_invalid_frames'] = 1
    item = self[target]

    # import xdev
    # xdev.profile.print_report()

    #print('item summary: ' + ub.urepr(self.summarize_item(item), nl=3))
    canvas1 = self.draw_item(item, overlay_on_image=0, rescale=1,
                             max_channels=5,
                             draw_truth=False,
                             draw_weights=False,
                             # combinable_extra='invariants.0:3'
                             )

    unmasked_target = target.copy()
    unmasked_target['mask_low_quality'] = 0
    item_unmasked = self[unmasked_target]
    canvas2 = self.draw_item(item_unmasked, overlay_on_image=0, rescale=1,
                             max_channels=1,
                             draw_truth=False,
                             draw_weights=False,
                             # combinable_extra='invariants.0:3'
                             )

    canvas = kwimage.stack_images([canvas1, canvas2], axis=0)

    # xdoctest: +REQUIRES(--show)
    import kwplot
    kwplot.autompl()
    kwplot.imshow(canvas, fnum=1)
    kwplot.show_if_requested()

    if 0:
        # resample the same target but no resampling
        target = item['target']
        target['resample_invalid_frames'] = 0
        target['force_bad_frames'] = 1
        target['mask_low_quality'] = 1
        # And different properties
        item = self[target]

        canvas = self.draw_item(item, overlay_on_image=0, rescale=1,
                                max_channels=5,
                                # combinable_extra='invariants.0:3'
                                )
        import kwplot
        kwplot.autompl()
        kwplot.imshow(canvas, fnum=2)
        kwplot.show_if_requested()

    if 0:
        from watch.tasks.fusion.datamodules.qa_bands import QA_SPECS
        import kwarray
        import ubelt as ub
        plt = kwplot.autoplt()
        # Pick a cloudmask that looks questionable
        chosen_index = 3

        frame_item = item['frames'][chosen_index]
        qa_data = frame_item['modes']['cloudmask'].numpy()
        print(ub.urepr(ub.udict(ub.dict_hist(qa_data.ravel())).sorted_keys()))
        # qa_data[np.isnan(qa_data)] = -9999
        # qa_data = np.clip(qa_data, 0, None)
        qa_data = qa_data.astype('int16')[0]

        # We don't have the exact right information here, so we can
        # punt for now and assume "Drop4"
        spec_name = 'ACC-1'
        gid = frame_item['gid']
        coco_img = self.sampler.dset.coco_image(gid)
        sensor = coco_img.img.get('sensor_coarse', '*')
        table = QA_SPECS.find_table(spec_name, sensor)
        print(f'table={table}')
        print('table.spec = {}'.format(ub.urepr(table.spec, nl=2)))

        fpath = ub.Path(coco_img.bundle_dpath) / coco_img.find_asset_obj('cloudmask')['file_name']
        qa_data = kwimage.imread(fpath)

        space = 'video'
        # space = 'image'
        scale = 0.4
        qa_data = coco_img.imdelay('cloudmask', space=space).scale(scale).finalize(interpolation='nearest', antialias=1)
        rgb_img = coco_img.imdelay('red|green|blue', space=space).scale(scale).finalize(interpolation='linear', nodata_method='ma')

        quality_im = qa_data
        result = table.draw_labels(quality_im)
        qa_canvas = result['qa_canvas']
        iffy_qa_names = [
            'cloud',
            # 'dilated_cloud',
            'cirrus',
        ]
        qa_names = iffy_qa_names
        is_cloud_iffy = table.mask_any(qa_data, qa_names)
        # norm_rgb, infos_to_return = kwarray.robust_normalize(rgb_img.clip(0, None), axis=None, params=dict(scaling='sigmoid', low=0.00, mid=0.8, high=0.99, extrema='quantile'), return_info=True)
        norm_rgb, infos_to_return = kwarray.robust_normalize(rgb_img.clip(0, None), axis=None, params=dict(scaling='sigmoid'), return_info=True)
        # norm_rgb = kwarray.robust_normalize(rgb_img)
        kwplot.imshow(norm_rgb, fnum=3, pnum=(1, 4, 1), title='rgb')
        kwplot.imshow(is_cloud_iffy, pnum=(1, 4, 2), title=f'mask matching {iffy_qa_names}')
        kwplot.imshow(qa_canvas, fnum=3, pnum=(1, 4, 3), title='qa bits')
        kwplot.imshow(result['legend'], fnum=3, pnum=(1, 4, 4), title='qa legend')
        kwplot.set_figtitle(f"QA Spec: name={table.spec['qa_spec_name']} sensor={table.spec['sensor']}")
        ax = plt.gca()
        ax.figure.tight_layout()


#     if 0:
#         data = item['frames'][0]['modes']['invariants.0|invariants.1|invariants.2'][..., 0].numpy()


# import timerit
# ti = timerit.Timerit(100, bestof=10, verbose=2)
# for timer in ti.reset('time'):
#     with timer:
#         # Speed up the compuation by doing this at a coarser scale
#         is_samecolor = util_kwimage.find_samecolor_regions(bands, scale=0.4, min_region_size=49)

# import timerit
# ti = timerit.Timerit(100, bestof=10, verbose=2)
# for timer in ti.reset('time'):
#     with timer:
#         util_kwimage.find_high_frequency_values(bands)


def _check_target(self):
    target = self.new_sample_grid['targets'][0]

    dset = self.sampler.dset
    gid_to_isbad = {}
    gid_to_sample = {}
    with_annots = 0
    coco_dset = dset
    sampler = self.sampler
    target_ = target.copy()
    gids = target_['gids']
    gid = gids[0]

    dset.images(target['gids']).lookup('sensor_coarse')

    item = self.getitem(target)

    coco_img = dset.coco_image(gid)

    import kwplot
    kwplot.autompl()
    kwplot.figure()

    kwplot.imshow(self.draw_item(item))

    data = coco_img.imdelay('invariants.0').finalize()
    kwplot.imshow(data)

    target_ = item['target'].copy()
    gid = target_['gids'][0]

    self._sample_one_frame(gid, sampler, coco_dset, target_, with_annots,
                           gid_to_isbad, gid_to_sample)
    sample = gid_to_sample[gid]

    for mode_name, mode in sample.items():
        pass

    # mode_name = 'invariants.0|invariants.1|invariants.2|invariants.3|invariants.4|invariants.5|invariants.6|invariants.7|invariants.8|invariants.9|invariants.10|invariants.11|invariants.12|invariants.13|invariants.14|invariants.15|invariants.16'



import os


TRUTHY_ENVIRONS = {'true', 'on', 'yes', '1', 't'}


def envflag(key, default=None, environ=None):
    """
    Determine if an environment variable is specified and truthy or falsy.

    Args:
        key (str): the environment variable name to check

        default (Any):
            the default value to return if the environment variable is not
            specified.

        environ (None | Dict):
            Uses this to get the environment variable. If unspecified, defaults
            to ``os.environ``.

    Returns:
        True if the environment variable exist and matches a truthy pattern.
        (e.g. true, on, yes, 1, or t). Otherwise returns False.

    Note:
        This will return false on any setting of the environ that is not
        truthy. (e.g. YESPLEASE is not a registered TRUTHY_ENVIRON so
        it will return False).

    Example:
        >>> from watch.utils import util_environ
        >>> environ = {
        >>>     'foo': '1',
        >>>     'bar': 'YES',
        >>>     'baz': '0',
        >>>     'biz': '1111',
        >>> }
        >>> assert util_environ.envflag('foo', 0, environ=environ)
        >>> assert util_environ.envflag('bar', 0, environ=environ)
        >>> assert not util_environ.envflag('baz', 0, environ=environ)
        >>> assert not util_environ.envflag('biz', 0, environ=environ)
        >>> assert not util_environ.envflag('buzz', 0, environ=environ)
        >>> assert util_environ.envflag('buzz', 1, environ=environ)
    """
    if environ is None:
        environ = os.environ
    if key not in environ:
        return default
    value = environ[key]
    if isinstance(value, str):
        value = value.lower() in TRUTHY_ENVIRONS
    else:
        assert value is None, 'environ values should all be strings'
    return value



r"""
python -m watch.tasks.fusion fit --help

Testing ddp

# Fit
export CUDA_VISIBLE_DEVICES=1
PHASE2_DATA_DPATH=$(geowatch_dvc --tags="phase2_data" --hardware="ssd")
PHASE2_EXPT_DPATH=$(geowatch_dvc --tags="phase2_expt")
DATASET_CODE=Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC
TRAIN_FNAME=combo_train_I.kwcoco.json
VALI_FNAME=combo_vali_I.kwcoco.json
TEST_FNAME=combo_vali_I.kwcoco.json
WORKDIR=$PHASE2_EXPT_DPATH/training/$HOSTNAME/$USER
KWCOCO_BUNDLE_DPATH=$PHASE2_DATA_DPATH/$DATASET_CODE
TRAIN_FPATH=$KWCOCO_BUNDLE_DPATH/$TRAIN_FNAME
VALI_FPATH=$KWCOCO_BUNDLE_DPATH/$VALI_FNAME
TEST_FPATH=$KWCOCO_BUNDLE_DPATH/$TEST_FNAME
INITIAL_STATE=noop
EXPERIMENT_NAME=Drop4_BAS_invariants_30GSD_V016
DEFAULT_ROOT_DIR=$WORKDIR/$DATASET_CODE/runs/$EXPERIMENT_NAME
python -m watch.tasks.fusion fit \
    --trainer.default_root_dir="$DEFAULT_ROOT_DIR" \
    --data.train_dataset="$TRAIN_FPATH" \
    --data.vali_dataset="$VALI_FPATH" \
    --data.time_steps=3 \
    --data.chip_size=96 \
    --data.batch_size=2 \
    --data.input_space_scale=30GSD \
    --data.window_space_scale=30GSD \
    --data.output_space_scale=30GSD \
    --model=watch.tasks.fusion.methods.HeterogeneousModel \
    --model.name="$EXPERIMENT_NAME" \
    --optimizer=torch.optim.AdamW \
    --optimizer.lr=1e-3 \
    --trainer.accelerator="gpu" \
    --trainer.devices="0," \
    --data.channels="red|green|blue,invariants:17"

"""


def test_noop_model_training():
    import pytest
    pytest.skip()
    from watch.tasks.fusion import fit_lightning
    import ubelt as ub
    config = ub.codeblock(
        '''
        subcommand: fit
        fit:
            model:
              class_path: NoopModel
            data:
              train_dataset: watch-msi-dates-geodata-gsize64-videos5-frames10
              vali_dataset: watch-msi-dates-geodata-gsize64-videos2-frames10
              chip_size: 64
              num_workers: 0
              batch_size: 1
              time_steps: 2
              channels: B1,B10,B11
            trainer:
              max_steps: 2
              num_sanity_val_steps: 0
        ''')
    fit_lightning.main(config=config)


def test_fit_cli_training():
    """
    python -m watch.tasks.fusion fit --model.help
    """
    import pytest
    pytest.skip()
    # Train a real model
    real_test_train()


def real_test_train():
    """
    xdoctest ~/code/watch/tests/test_lightning_cli_fit.py real_test_train
    python -c "from test_lightning_cli_fit import real_test_train; real_test_train()"
    from test_lightning_cli_fit import real_test_train
    real_test_train()
    """
    from watch.tasks.fusion import fit_lightning
    import ubelt as ub
    dpath = ub.Path.appdir('watch/tests/test_fusion_fit/demo_real').ensuredir()
    config = ub.codeblock(
        f'''
        subcommand: fit
        fit:
            data:
              train_dataset: watch-msi-dates-geodata-gsize64-videos5-frames10
              vali_dataset: watch-msi-dates-geodata-gsize64-videos2-frames10
              num_workers: 8
              batch_size: 16
              time_steps: 2
              channels: B1,B10,B11
              chip_overlap: 0.2
              chip_size: 64
              dist_weights: 0
              temporal_dropout: 0.0
              time_sampling: contiguous
              time_span: 2y
              verbose: 1
            model:
              class_path: watch.tasks.fusion.methods.HeterogeneousModel
              init_args:
                token_width: 8
                token_dim: 64
                position_encoder:
                  class_path: watch.tasks.fusion.methods.heterogeneous.MipNerfPositionalEncoder
                  init_args:
                    in_dims: 3
                    max_freq: 3
                    num_freqs: 16
                backbone:
                  class_path: watch.tasks.fusion.architectures.transformer.TransformerEncoderDecoder
                  init_args:
                    encoder_depth: 2
                    decoder_depth: 2
                    dim: 160
                    queries_dim: 96
                    logits_dim: 64
                    latent_dim_head: 256
                spatial_scale_base: 1.0
                temporal_scale_base: 1.0
                global_change_weight: 0.0
                global_class_weight: 1.0
                global_saliency_weight: 0.5
                class_loss: dicefocal
                saliency_loss: focal
                decoder: simple_conv
            optimizer:
              class_path: torch.optim.AdamW
              init_args:
                lr: 1e-4
                weight_decay: 0
            profile: false
            seed_everything: true
            trainer:
              default_root_dir: {dpath}
              # max_steps: 2000
              logger: true
              # accelerator: gpu
              # devices: 1
              # callbacks:
              #   - class_path: pytorch_lightning.callbacks.ModelCheckpoint
              #     init_args:
              #       monitor: val_class_f1_macro
              #       save_top_k: 10
              #       mode: max
              # check_val_every_n_epoch: 5
              # enable_checkpointing: true
              # enable_model_summary: true
              # enable_progress_bar: true
              # gradient_clip_algorithm: null
              # gradient_clip_val: null
              # log_every_n_steps: 50
              # num_sanity_val_steps: 2
              # replace_sampler_ddp: true
              # track_grad_norm: 2

        ''')

    fit_lightning.main(config)


def test_partial_init_callback():
    from watch.tasks.fusion import fit_lightning
    import ubelt as ub

    import pytest
    pytest.skip('todo: make this test run faster')

    dpath1 = ub.Path.appdir('watch/tests/test_fusion_fit/partial_init/base_model').ensuredir()
    dpath1.delete().ensuredir()
    # Get the package we just trained and init from it
    # avail_package_fpaths = (sorted((dpath1 / 'lightning_logs/').glob('*'))

    # a = (sorted((dpath1 / 'lightning_logs/').glob('*')))
    # if len(a):
    #     avail_package_fpaths = (sorted(a[-1] / 'packages').glob('*.pt'))
    # else:
    #     avail_package_fpaths = []
    # if len(avail_package_fpaths) == 0:
    # Make an initializer package only if we need to.
    config = ub.codeblock(
        f'''
        subcommand: fit
        fit:
            model:
              class_path: watch.tasks.fusion.methods.HeterogeneousModel
            data:
              train_dataset: watch-msi-dates-geodata-gsize64-videos5-frames10
              vali_dataset: watch-msi-dates-geodata-gsize64-videos2-frames10
              num_workers: 0
              batch_size: 1
              time_steps: 2
              channels: B1,B10,B11
            optimizer:
              class_path: torch.optim.SGD
              init_args:
                lr: 1e-3
            trainer:
              default_root_dir: {dpath1}
              max_steps: 2
              num_sanity_val_steps: 0
        ''')
    fit_lightning.main(config=config)

    avail_package_fpaths = sorted(dpath1.glob('*.pt'))
    if len(avail_package_fpaths) == 0:
        raise AssertionError('We should have produced a trained model')

    package_fpath = avail_package_fpaths[-1]

    dpath2 = ub.Path.appdir('watch/tests/test_fusion_fit/partial_init/preinit_model')
    dpath2.delete().ensuredir()
    config = ub.codeblock(
        f'''
        subcommand: fit
        fit:
            model:
              class_path: watch.tasks.fusion.methods.HeterogeneousModel
            data:
              train_dataset: watch-msi-dates-geodata-gsize64-videos5-frames10
              vali_dataset: watch-msi-dates-geodata-gsize64-videos2-frames10
              num_workers: 0
              batch_size: 1
              time_steps: 2
              channels: B1,B11,r|g|b
            optimizer:
              class_path: torch.optim.SGD
              init_args:
                lr: 1e-3
            trainer:
              default_root_dir: {dpath2}
              max_steps: 2
              num_sanity_val_steps: 0
            initializer:
              init: {package_fpath}
        ''')
    cli = fit_lightning.main(config=config)
    print(f'cli={cli}')
    dpath1.delete()
    dpath2.delete()


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/tests/test_lightning_cli_fit.py
    """
    test_partial_init_callback()



# Autogenerated via:
# python ~/code/watch/dev/maintain/mirror_package_geowatch.py
def __getattr__(key):
    import watch.tasks.fusion.predict as mirror
    return getattr(mirror, key)


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/geowatch/tasks/fusion/predict.py
    """
    from watch.tasks.fusion.predict import main
    main()



#!/usr/bin/env python3
# flake8: noqa
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import argparse
import builtins
import math
import os
import random
import shutil
import time
import warnings

import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import torch.optim
import torch.multiprocessing as mp
import torch.utils.data
import torch.utils.data.distributed
import torchvision.transforms as transforms
import torchvision.datasets as datasets
import torchvision.models as models

import watch.tasks.rutgers_material_seg.models.moco as moco
# import moco.builder

model_names = sorted(name for name in models.__dict__
                     if name.islower() and not name.startswith("__")
                     and callable(models.__dict__[name]))

parser = argparse.ArgumentParser(description='PyTorch ImageNet Training')
parser.add_argument('data', metavar='DIR',
                    help='path to dataset')
parser.add_argument('-a', '--arch', metavar='ARCH', default='resnet50',
                    choices=model_names,
                    help='model architecture: ' +
                        ' | '.join(model_names) +
                        ' (default: resnet50)')
parser.add_argument('-j', '--workers', default=32, type=int, metavar='N',
                    help='number of data loading workers (default: 32)')
parser.add_argument('--epochs', default=200, type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('--start-epoch', default=0, type=int, metavar='N',
                    help='manual epoch number (useful on restarts)')
parser.add_argument('-b', '--batch-size', default=256, type=int,
                    metavar='N',
                    help='mini-batch size (default: 256), this is the total '
                         'batch size of all GPUs on the current node when '
                         'using Data Parallel or Distributed Data Parallel')
parser.add_argument('--lr', '--learning-rate', default=0.03, type=float,
                    metavar='LR', help='initial learning rate', dest='lr')
parser.add_argument('--schedule', default=[120, 160], nargs='*', type=int,
                    help='learning rate schedule (when to drop lr by 10x)')
parser.add_argument('--momentum', default=0.9, type=float, metavar='M',
                    help='momentum of SGD solver')
parser.add_argument('--wd', '--weight-decay', default=1e-4, type=float,
                    metavar='W', help='weight decay (default: 1e-4)',
                    dest='weight_decay')
parser.add_argument('-p', '--print-freq', default=10, type=int,
                    metavar='N', help='print frequency (default: 10)')
parser.add_argument('--resume', default='', type=str, metavar='PATH',
                    help='path to latest checkpoint (default: none)')
parser.add_argument('--world-size', default=-1, type=int,
                    help='number of nodes for distributed training')
parser.add_argument('--rank', default=-1, type=int,
                    help='node rank for distributed training')
parser.add_argument('--dist-url', default='tcp://224.66.41.62:23456', type=str,
                    help='url used to set up distributed training')
parser.add_argument('--dist-backend', default='nccl', type=str,
                    help='distributed backend')
parser.add_argument('--seed', default=None, type=int,
                    help='seed for initializing training. ')
parser.add_argument('--gpu', default=None, type=int,
                    help='GPU id to use.')
parser.add_argument('--multiprocessing-distributed', action='store_true',
                    help='Use multi-processing distributed training to launch '
                         'N processes per node, which has N GPUs. This is the '
                         'fastest way to use PyTorch for either single node or '
                         'multi node data parallel training')

# moco specific configs:
parser.add_argument('--moco-dim', default=128, type=int,
                    help='feature dimension (default: 128)')
parser.add_argument('--moco-k', default=65536, type=int,
                    help='queue size; number of negative keys (default: 65536)')
parser.add_argument('--moco-m', default=0.999, type=float,
                    help='moco momentum of updating key encoder (default: 0.999)')
parser.add_argument('--moco-t', default=0.07, type=float,
                    help='softmax temperature (default: 0.07)')

# options for moco v2
parser.add_argument('--mlp', action='store_true',
                    help='use mlp head')
parser.add_argument('--aug-plus', action='store_true',
                    help='use moco v2 data augmentation')
parser.add_argument('--cos', action='store_true',
                    help='use cosine lr schedule')


def main():
    args = parser.parse_args()

    if args.seed is not None:
        random.seed(args.seed)
        torch.manual_seed(args.seed)
        cudnn.deterministic = True
        warnings.warn('You have chosen to seed training. '
                      'This will turn on the CUDNN deterministic setting, '
                      'which can slow down your training considerably! '
                      'You may see unexpected behavior when restarting '
                      'from checkpoints.')

    if args.gpu is not None:
        warnings.warn('You have chosen a specific GPU. This will completely '
                      'disable data parallelism.')

    if args.dist_url == "env://" and args.world_size == -1:
        args.world_size = int(os.environ["WORLD_SIZE"])

    args.distributed = args.world_size > 1 or args.multiprocessing_distributed

    ngpus_per_node = torch.cuda.device_count()
    if args.multiprocessing_distributed:
        # Since we have ngpus_per_node processes per node, the total world_size
        # needs to be adjusted accordingly
        args.world_size = ngpus_per_node * args.world_size
        # Use torch.multiprocessing.spawn to launch distributed processes: the
        # main_worker process function
        mp.spawn(main_worker, nprocs=ngpus_per_node, args=(ngpus_per_node, args))
    else:
        # Simply call main_worker function
        main_worker(args.gpu, ngpus_per_node, args)


def main_worker(gpu, ngpus_per_node, args):
    args.gpu = gpu

    # suppress printing if not master
    if args.multiprocessing_distributed and args.gpu != 0:
        def print_pass(*args):
            pass
        builtins.print = print_pass

    if args.gpu is not None:
        print("Use GPU: {} for training".format(args.gpu))

    if args.distributed:
        if args.dist_url == "env://" and args.rank == -1:
            args.rank = int(os.environ["RANK"])
        if args.multiprocessing_distributed:
            # For multiprocessing distributed training, rank needs to be the
            # global rank among all the processes
            args.rank = args.rank * ngpus_per_node + gpu
        dist.init_process_group(backend=args.dist_backend, init_method=args.dist_url,
                                world_size=args.world_size, rank=args.rank)
    # create model
    print("=> creating model '{}'".format(args.arch))
    model = moco.builder.MoCo(
        models.__dict__[args.arch],
        args.moco_dim, args.moco_k, args.moco_m, args.moco_t, args.mlp)
    print(model)

    if args.distributed:
        # For multiprocessing distributed, DistributedDataParallel constructor
        # should always set the single device scope, otherwise,
        # DistributedDataParallel will use all available devices.
        if args.gpu is not None:
            torch.cuda.set_device(args.gpu)
            model.cuda(args.gpu)
            # When using a single GPU per process and per
            # DistributedDataParallel, we need to divide the batch size
            # ourselves based on the total number of GPUs we have
            args.batch_size = int(args.batch_size / ngpus_per_node)
            args.workers = int((args.workers + ngpus_per_node - 1) / ngpus_per_node)
            model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.gpu])
        else:
            model.cuda()
            # DistributedDataParallel will divide and allocate batch_size to all
            # available GPUs if device_ids are not set
            model = torch.nn.parallel.DistributedDataParallel(model)
    elif args.gpu is not None:
        torch.cuda.set_device(args.gpu)
        model = model.cuda(args.gpu)
        # comment out the following line for debugging
        raise NotImplementedError("Only DistributedDataParallel is supported.")
    else:
        # AllGather implementation (batch shuffle, queue update, etc.) in
        # this code only supports DistributedDataParallel.
        raise NotImplementedError("Only DistributedDataParallel is supported.")

    # define loss function (criterion) and optimizer
    criterion = nn.CrossEntropyLoss().cuda(args.gpu)

    optimizer = torch.optim.SGD(model.parameters(), args.lr,
                                momentum=args.momentum,
                                weight_decay=args.weight_decay)

    # optionally resume from a checkpoint
    if args.resume:
        if os.path.isfile(args.resume):
            print("=> loading checkpoint '{}'".format(args.resume))
            if args.gpu is None:
                checkpoint = torch.load(args.resume)
            else:
                # Map model to be loaded to specified single gpu.
                loc = 'cuda:{}'.format(args.gpu)
                checkpoint = torch.load(args.resume, map_location=loc)
            args.start_epoch = checkpoint['epoch']
            model.load_state_dict(checkpoint['state_dict'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            print("=> loaded checkpoint '{}' (epoch {})"
                  .format(args.resume, checkpoint['epoch']))
        else:
            print("=> no checkpoint found at '{}'".format(args.resume))

    cudnn.benchmark = True

    # Data loading code
    traindir = os.path.join(args.data, 'train')
    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])
    if args.aug_plus:
        # MoCo v2's aug: similar to SimCLR https://arxiv.org/abs/2002.05709
        augmentation = [
            transforms.RandomResizedCrop(224, scale=(0.2, 1.)),
            transforms.RandomApply([
                transforms.ColorJitter(0.4, 0.4, 0.4, 0.1)  # not strengthened
            ], p=0.8),
            transforms.RandomGrayscale(p=0.2),
            transforms.RandomApply([moco.loader.GaussianBlur([.1, 2.])], p=0.5),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            normalize
        ]
    else:
        # MoCo v1's aug: the same as InstDisc https://arxiv.org/abs/1805.01978
        augmentation = [
            transforms.RandomResizedCrop(224, scale=(0.2, 1.)),
            transforms.RandomGrayscale(p=0.2),
            transforms.ColorJitter(0.4, 0.4, 0.4, 0.4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            normalize
        ]

    # train_dataset = datasets.ImageFolder(
    #     traindir,
    #     moco.loader.TwoCropsTransform(transforms.Compose(augmentation)))

    train_dataset = datasets.ImageFolder(
        traindir,
        moco.loader.TwoCropsTransform(transforms.Compose(augmentation)))

    if args.distributed:
        train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset)
    else:
        train_sampler = None

    train_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=args.batch_size, shuffle=(train_sampler is None),
        num_workers=args.workers, pin_memory=True, sampler=train_sampler, drop_last=True)

    for epoch in range(args.start_epoch, args.epochs):
        if args.distributed:
            train_sampler.set_epoch(epoch)
        adjust_learning_rate(optimizer, epoch, args)

        # train for one epoch
        train(train_loader, model, criterion, optimizer, epoch, args)

        if not args.multiprocessing_distributed or (args.multiprocessing_distributed
                and args.rank % ngpus_per_node == 0):
            save_checkpoint({
                'epoch': epoch + 1,
                'arch': args.arch,
                'state_dict': model.state_dict(),
                'optimizer' : optimizer.state_dict(),
            }, is_best=False, filename='checkpoint_{:04d}.pth.tar'.format(epoch))


def train(train_loader, model, criterion, optimizer, epoch, args):
    batch_time = AverageMeter('Time', ':6.3f')
    data_time = AverageMeter('Data', ':6.3f')
    losses = AverageMeter('Loss', ':.4e')
    top1 = AverageMeter('Acc@1', ':6.2f')
    top5 = AverageMeter('Acc@5', ':6.2f')
    progress = ProgressMeter(
        len(train_loader),
        [batch_time, data_time, losses, top1, top5],
        prefix="Epoch: [{}]".format(epoch))

    # switch to train mode
    model.train()

    end = time.time()
    for i, (images, _) in enumerate(train_loader):
        # measure data loading time
        data_time.update(time.time() - end)

        if args.gpu is not None:
            images[0] = images[0].cuda(args.gpu, non_blocking=True)
            images[1] = images[1].cuda(args.gpu, non_blocking=True)

        # compute output
        output, target = model(im_q=images[0], im_k=images[1])

        loss = criterion(output, target)

        # acc1/acc5 are (K+1)-way contrast classifier accuracy
        # measure accuracy and record loss
        acc1, acc5 = accuracy(output, target, topk=(1, 5))
        losses.update(loss.item(), images[0].size(0))
        top1.update(acc1[0], images[0].size(0))
        top5.update(acc5[0], images[0].size(0))

        # compute gradient and do SGD step
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % args.print_freq == 0:
            progress.display(i)


def save_checkpoint(state, is_best, filename='checkpoint.pth.tar'):
    torch.save(state, filename)
    if is_best:
        shutil.copyfile(filename, 'model_best.pth.tar')


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, name, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)


class ProgressMeter(object):
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        print('\t'.join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = '{:' + str(num_digits) + 'd}'
        return '[' + fmt + '/' + fmt.format(num_batches) + ']'


def adjust_learning_rate(optimizer, epoch, args):
    """Decay the learning rate based on schedule"""
    lr = args.lr
    if args.cos:  # cosine lr schedule
        lr *= 0.5 * (1. + math.cos(math.pi * epoch / args.epochs))
    else:  # stepwise lr schedule
        for milestone in args.schedule:
            lr *= 0.1 if epoch >= milestone else 1.
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res


if __name__ == '__main__':
    main()
